import os
import pandas as pd 
import numpy as np

from ms_mint.standards import MINT_ROOT
from ms_mint import processing


TEST_MZXML = os.path.abspath( 
    os.path.join( 
        MINT_ROOT ,'..' ,'tests' , 'data', 'test.mzXML'
    )
)

assert os.path.isfile(TEST_MZXML)    


def test__find_test_mzxml():
    assert os.path.isfile(TEST_MZXML),\
        'Test mzXML ({}) not found.'.format(TEST_MZXML)


def test__process_ms1():
    ms_data = pd.DataFrame({'retentionTime': [1, 2, 3],
                            'm/z array': [100, 200, 300], 
                            'intensity array': [2, 3, 7]})

    peaklist = pd.DataFrame(
        {'peak_label': ['A'],
         'mz_mean': [200],
         'mz_width': [10],
         'intensity_threshold': [0],
         'rt_min': [0],
         'rt_max': [10]})

    result = processing.process_ms1(ms_data, peaklist)

    expect= pd.DataFrame({'peak_label': {0: 'A'},
                             'mz_mean': {0: 200},
                             'mz_width': {0: 10},
                             'intensity_threshold': {0: 0},
                             'rt_min': {0: 0},
                             'rt_max': {0: 10},
                             'peak_area': {0: 3},
                             'peak_n_datapoints': {0: 1},
                             'peak_max': {0: 3},
                             'peak_rt_of_max': {0: 2},
                             'peak_min': {0: 3},
                             'peak_median': {0: 3.0},
                             'peak_mean': {0: 3.0},
                             'peak_delta_int': {0: 0},
                             'peak_shape_rt': {0: '2'},
                             'peak_shape_int': {0: '3'},
                             'peak_mass_diff_25pc': {0: 0.0},
                             'peak_mass_diff_50pc': {0: 0.0},
                             'peak_mass_diff_75pc': {0: 0.0}})
    print(result.values)
    print(expect.values)
    print(result.columns.to_list())
    assert result.equals(expect), result


def test__process_ms1_from_df():
    df = pd.DataFrame({'retentionTime': [1, 2, 3],
                       'm/z array': [100,200,300], 
                       'intensity array': [2, 3, 7]})
    peaklist = pd.DataFrame(
        {'peak_label': ['A'],
         'mz_mean': [200],
         'mz_width': [10],
         'intensity_threshold': [0],
         'rt_min': [0],
         'rt_max': [10]})
    result = processing.process_ms1_from_df(df, peaklist)
    expect= [['A', 3, 1, 3, 2, 3, 3.0, 3.0, 
                 0, '2', '3', 0.0, 0.0, 0.0]]
    print(result)
    print(expect) 
    assert result == expect


def test__slice_ms1_array():
    array = np.array([[  1, 100,   2],
                      [  2, 200,   3],
                      [  3, 300,   7]])
    result = processing.slice_ms1_array(array,  rt_min=1.5, rt_max=2.5, 
        mz_mean=200, mz_width=10, intensity_threshold=0)
    expect = np.array([[  2, 200,   3]])
    print(expect)
    print(result)
    assert np.array_equal(result, expect)


def test__process_ms1_from_numpy():
    array = np.array([[  1, 100,   2],
                      [  2, 200,   3],
                      [  3, 300,   7]])

    peaks = [(100, 10, .5, 1.5, 0, 'A'), 
             (200, 10, 1.5, 2.5, 0, 'B')]

    result = processing.process_ms1_from_numpy(array, peaks)

    expect = [['A', 2, 1, 2, 1, 2, 2.0, 2.0, 0, '1', '2', 0.0, 0.0, 0.0],
              ['B', 3, 1, 3, 2, 3, 3.0, 3.0, 0, '2', '3', 0.0, 0.0, 0.0]]
    
    print('Expected:')
    [print(i) for i in expect]
    print('Actual')
    [print(i) for i in result]
    assert result == expect


def button_style(style_name=None, **kwargs):
        style = {'color': 'black', 
                 'margin-right': 5, 
                 'margin-bottom': '1.5em',
                 'background-color': 'lightgrey'}
        
        styles = {'warn':    {'background-color': 'orange',
                              'font-weight': 'bold', 
                              'color': 'black'},
                  'error':   {'background-color': 'orange'},
                  'wait':    {'color': 'grey'},
                  'ready':   {'background-color': 'lightgreen',
                              'color': 'black'},
                  'help':    {'background-color': 'lightblue'},
                  'next':    {'font-weight': 'bold', 
                              'border': '2px solid blue'},
                  'neutral': {'background-color': 'white'}}
        
        if style_name in styles.keys():
            style.update(styles[style_name])
            
        style.update(**kwargs)

        return style



import dash_html_components as html
import dash_core_components as dcc

from dash_table import DataTable


ms_table = html.Div(id='ms-table-container', 
    style={'min-height':  100, 'margin': '5%'},
    children=[
        DataTable(id='ms-table',
            columns=[ {"name": i, "id": i, "selectable": True, 'editable': i != 'MS-file'}  
                                for i in ['MS-file', 'Label']],
                    data=None,
                    row_selectable='multi',
                    row_deletable=False,
                    style_cell={'textAlign': 'left'},
                    sort_action='native',
                    filter_action='native',                
                    )
])

ms_layout = html.Div([
    html.H3('MS-file'),
    dcc.Upload(
            id='ms-upload',
            children=html.Div([
                'Drag and Drop or ',
                html.A('Select Files')
            ]),
            style={
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin': '10px'
            },
            # Allow multiple files to be uploaded
            multiple=True
        ),
    html.Button('Convert to Feather', id='ms-convert'),
    dcc.Loading( html.Div(id='ms-upload-output') ),
    html.Div(id='ms-convert-output'),
    html.Div(id='ms-delete-output'),
    dcc.Loading( ms_table ), 
    html.Div(id='ms-save-output'),
    dcc.Input(id='ms-input'),
    html.Button('Set Labels', id='ms-set-labels'),
    #html.Button('Set Batch', id='ms-set-batch'),
    html.Button('Delete selected files', id='ms-delete', style={'float': 'right'}),
])


# ms_mint/io.py

import pandas as pd
import numpy as np
import io
import os
import pymzml

from datetime import date
from pyteomics import mzxml, mzml


def ms_file_to_df(fn):
    if fn.lower().endswith('.mzxml'):
        return mzxml_to_pandas_df(fn)
    elif fn.lower().endswith('.mzml'):
        return mzml_to_df(fn)
    elif fn.lower().endswith('hdf'):
        return pd.read_hdf(fn)
    elif fn.lower().endswith('feather'):
        return pd.read_feather(fn)


def mzxml_to_pandas_df(fn):
    '''
    Reads mzXML file and returns a pandas.DataFrame.
    '''
    cols = ['retentionTime', 'm/z array', 'intensity array']
    slices = []
    file = mzxml.MzXML(fn)
    while True:
        try:
            data = file.next()     
            df = pd.DataFrame({col: np.array(data[col]) for col in cols} )
            slices.append( df )
        except:         
            break  
    df = pd.concat(slices)
    df['retentionTime'] =  df['retentionTime'].astype(np.float32)
    df['m/z array'] = df['m/z array'].astype(np.float32)
    df['intensity array'] = df['intensity array'].astype(int)
    df = df.reset_index(drop=True)
    return df


def mzml_to_pandas_df(fn):
    '''
    Reads mzML file and returns a pandas.DataFrame.
    '''
    cols = ['retentionTime', 'm/z array', 'intensity array']
    slices = []
    file = mzml.MzML(fn)
    while True:
        try:
            data = file.next()
            data['retentionTime'] = data['scanList']['scan'][0]['scan time'] / 60
            del data['scanList']
            slices.append( pd.DataFrame(data) ) 
        except:
            break
    df = pd.concat(slices)[cols]
    df_to_numeric(df)
    df['intensity array'] = df['intensity array'].astype(int)
    df = df.reset_index(drop=True)
    return df


def mzml_to_df(fn, assume_time_unit='seconds'):
    run = pymzml.run.Reader(fn)
    data = []
    for spectrum in run:
        # Try to convert time units with build-in method
        # some files have no time unit set. Then convert 
        # to minutes assuming the time unit is as set
        # by assume_time_unit argument.
        try:
            RT = spectrum.scan_time_in_minutes()
        except:
            if assume_time_unit == 'seconds':
                RT = spectrum.scan_time[0] / 60.
            elif assume_time_unit == 'minutes':
                RT = spectrum.scan_time[0]
            
        peaks = spectrum.peaks("centroided")
        data.append((RT,peaks))

    df = pd.DataFrame(data).explode(1)

    df['m/z array'] = df[1].apply(lambda x: x[0])
    df['intensity array'] = df[1].apply(lambda x: x[1]).astype(int)
    df = df.rename(columns={0: 'retentionTime'})
    del df[1]
    df = df.reset_index(drop=True)
    return df


def df_to_numeric(df):
    '''
    Converts dataframe to numeric types if possible.
    '''
    for col in df.columns:
        df.loc[:, col] = pd.to_numeric(df[col], errors='ignore')


def export_to_excel(mint, fn=None):
    date_string = str(date.today())
    if fn is None:
        file_buffer = io.BytesIO()
        writer = pd.ExcelWriter(file_buffer)
    else:
        writer = pd.ExcelWriter(fn)
    # Write into file
    mint.peaklist.to_excel(writer, 'Peaklist', index=False)
    mint.results.to_excel(writer, 'Results', index=False)
    meta = pd.DataFrame({'MINT_version': [mint.version], 
                         'Date': [date_string]}).T[0]
    meta.to_excel(writer, 'Metadata', index=True, header=False)
    # Close writer and maybe return file buffer
    writer.close()
    if fn is None:
        return file_buffer.seek(0)


def convert_ms_file_to_feather(fn, fn_out=None):
    base, ext = os.path.splitext(fn)
    if fn_out is None:
        fn_out = base+'.feather'
    ms_file_to_df(fn).reset_index(drop=True).to_feather(fn_out)
    return fn_out


import numpy as np
from matplotlib import pyplot as plt
from scipy.cluster.hierarchy import linkage, dendrogram
from scipy.spatial.distance import pdist, squareform


def hierarchical_clustering(df, vmin=None, vmax=None, figsize=(8,8), top_height=2, left_width=2,
                            xmaxticks=None, ymaxticks=None, metric='euclidean', cmap=None):
    '''based on heatmap function from
    http://nbviewer.ipython.org/github/herrfz/dataanalysis/
    blob/master/week3/svd_pca.ipynb
    Generates a heatmap from the input matrix.
    '''

    df = df.copy()

    # Subplot sizes
    total_width, total_height = figsize 
    
    main_h = 1-(top_height/total_height)
    main_w = 1-(left_width/total_width)
    
    gap_x = (0.1/total_width)
    gap_y = (0.1/total_height) 
    
    left_h = main_h
    left_w = 1-main_w

    top_h = 1-main_h
    top_w = main_w
    

    ydim, xdim = df.shape

    if xmaxticks is None:
        xmaxticks = int( 5*main_w*total_width )
    if ymaxticks is None:
        ymaxticks = int( 5*main_h*total_height )
    
    dm = df.fillna(0).values
    D1 = squareform(pdist(dm, metric=metric))
    D2 = squareform(pdist(dm.T, metric=metric))
    fig = plt.figure(figsize=figsize)
    fig.set_tight_layout(False)

    
    # add left dendrogram
    ax1 = fig.add_axes([0, 0, left_w-gap_x, left_h], frameon=False)
    Y = linkage(D1, method='complete')
    Z1 = dendrogram(Y, orientation='left', color_threshold=0, above_threshold_color='k')
    ax1.set_xticks([])
    ax1.set_yticks([])
    # add top dendrogram
    ax2 = fig.add_axes([left_w, main_h+gap_y, top_w, top_h-gap_y], frameon=False)
    Y = linkage(D2, method='complete')
    Z2 = dendrogram(Y, color_threshold=0, above_threshold_color='k')
    ax2.set_xticks([])
    ax2.set_yticks([])
    # add matrix plot
    axmatrix = fig.add_axes([left_w, 0, main_w, main_h])
    idx1 = Z1['leaves']
    idx2 = Z2['leaves']
    D = dm[idx1, :]
    D = D[:, idx2]

    if cmap is None: cmap='coolwarm'
    fig = axmatrix.matshow(D[::-1], aspect='auto', cmap=cmap,
                           vmin=vmin, vmax=vmax)

    axmatrix.set_xticks([])
    axmatrix.set_yticks([])
    
    ax = plt.gca()
    ax.yaxis.tick_right()
    ax.xaxis.tick_bottom()

    clustered = df.iloc[Z1['leaves'][::-1], Z2['leaves']]

    ndx_y = np.linspace(0,len(clustered.index)-1, ymaxticks)
    ndx_x = np.linspace(0,len(clustered.columns)-1, xmaxticks)
    ndx_y = [int(i) for i in ndx_y]
    ndx_x = [int(i) for i in ndx_x]

    _ = plt.yticks(ndx_y, clustered.iloc[ndx_y].index)
    _ = plt.xticks(ndx_x, clustered.columns[ndx_x], rotation=90)

    return clustered, fig, Z1['leaves'][::-1], Z2['leaves']


# ms_mint/helpers.py


def dataframe_difference(df1, df2, which=None):
    """Find rows which are different between two DataFrames."""
    comparison_df = df1.merge(df2,
                              indicator=True,
                              how='outer')
    if which is None:
        diff_df = comparison_df[comparison_df['_merge'] != 'both']
    else:
        diff_df = comparison_df[comparison_df['_merge'] == which]
    return diff_df


def sort_columns_by_median(df):
    cols = df.median().sort_values(ascending=False).index
    return df[cols]


def remove_all_zero_columns(df):
    is_zero = df.max() != 0
    is_zero = is_zero[is_zero].index
    return df[is_zero]


def is_ms_file(fn):
    if (fn.lower().endswith('.mzxml')) or \
       (fn.lower().endswith('.mzml'))  or \
       (fn.lower().endswith('.mzhdf')) or \
       (fn.lower().endswith('.feather')):
        return True
    else:
        return False


import os, io
import ipywidgets as widgets
import tempfile

from glob import glob
from pathlib import Path as P 

from ipywidgets import Button, HBox, VBox, Textarea, Layout, FileUpload, Tab
from ipywidgets import IntProgress as Progress
from ipyfilechooser import FileChooser
from IPython.display import display
from IPython.core.display import HTML

from .Mint import Mint as MintBase


HOME = str(P.home())

class Mint(MintBase):
    def __init__(self, *args, **kwargs):

        self.progress_callback = self.set_progress

        super().__init__(progress_callback=self.progress_callback, *args, **kwargs)
        
        fc = FileChooser()
        fc.show_only_dirs = True
        fc.default_path = tempfile.gettempdir()
        
        self.ms_storage_path = fc

        self.ms_upload = FileUpload()

        self.peaklist_files_button = FileUpload(description='Peaklists', accept='csv,xlsx', multiple=False)

        self.peaklist_files_button.observe(self.load_peaklist, names='value')

        self.load_ms_button = Button(description='Load MS-files')

        self.load_ms_button.on_click(self.search_files)

        self.detect_peaks_button = Button(description="Detect Peaks")
        self.detect_peaks_button.on_click(self.detect_peaks)

        self.message_box = Textarea(
            value='',
            placeholder='Please select some files and click on Run.',
            description='',
            disabled=True,
            layout={'width': '90%', 
                    'height': '500px', 
                    'font_family': 'monospace'})
        
        self.run_button = Button(description="Run")
        self.run_button.on_click(self.run)
        self.run_button.style.button_color = 'lightgray'

        self.optimize_rt_button = Button(description="Optimize RT")
        self.optimize_rt_button.on_click(self.action_optimize_rt)
        
        self.download_button = Button(description="Export")
        self.download_button.on_click(self.export_action)
        self.download_button.style.button_color = 'lightgray'

        self.progress_bar = Progress(min=0, max=100, layout=Layout(width='90%'), 
                                 description='Progress:', bar_style='info')

        self.output = widgets.Output()

        tabs = Tab()
        tabs.children = [
                         HBox([self.ms_storage_path, self.ms_upload, self.load_ms_button]),
                         HBox([self.peaklist_files_button, self.detect_peaks_button, self.optimize_rt_button]),
                         ]

        tabs.set_title(0, 'MS-Files')
        tabs.set_title(1, 'Peaklists')

        self.layout = VBox([                      
                      tabs,
                            self.message_box,
                      HBox([self.run_button, 
                            self.download_button]),
                            self.progress_bar   
                ])
    
    def load_peaklist(self, value):
        for fn, data in value['new'].items():
            self.load(io.BytesIO(data['content']))
        self.list_files()

    def action_optimize_rt(self, b):
        if (self.n_files > 0) and len(self.peaklist)>0:
            self.optimize_retention_times()
        
    def message(self, text):
        self.message_box.value = f'{text}\n' + self.message_box.value

    def clear_messages(self):
        self.message_box.value = ''

    def search_files(self, b=None):
        self.ms_files = (glob(os.path.join(self.ms_storage_path.selected_path, '*mzXML')) +
                         glob(os.path.join(self.ms_storage_path.selected_path, '*mzML')) +
                         glob(os.path.join(self.ms_storage_path.selected_path, '*mzHDF')) + 
                         glob(os.path.join(self.ms_storage_path.selected_path, '*mzxml')) +
                         glob(os.path.join(self.ms_storage_path.selected_path, '*mzml')) +
                         glob(os.path.join(self.ms_storage_path.selected_path, '*mzhdf')) )
        self.list_files()

    def show(self):
        display(HTML("<style>textarea, input { font-family: monospace; }</style>"))
        return self.layout
            
    def files(self, files):
        super(Mint, self).ms_files = files
        self.ms_files_button.files = files
        self.list_files()

    def add_ms_files(self, fns):
        if fns is not None:
            self.ms_files = self.ms_files + fns
        self.list_files()

    def add_peaklist_files(self, fns):
        if fns is not None:
            self.peaklist_files = self.peaklist_files + fns
        self.list_files()

    def list_files(self, b=None):
        text = f'{self.n_files} MS-files to process:\n'
        for i, line in enumerate(self.ms_files):
            text += line+'\n'
            if i > 10:
                text += line+'\n...\n'
                break
        text += '\nUsing peak list:\n'
        if len(self.peaklist_files) != 0:
            text += '\n'.join([str(i) for i in self.peaklist_files])
        elif len(self.peaklist) != 0:
            text += self.peaklist.to_string()
        else:
            text += '\nNo peaklist defined.'
        self.message(text)
        if (self.n_files != 0) and (self.n_peaklist_files != 0):
            self.run_button.style.button_color = 'lightgreen'
        else:
            self.run_button.style.button_color = 'lightgray'
       
    def run(self, b=None, **kwargs):
        self.progress = 0
        super(Mint, self).run(**kwargs)
        self.message('Done processing MS-files.')
        if self.results is not None:
            self.download_button.style.button_color = 'lightgreen'
    
    def detect_peaks(self, b=None, **kwargs):
        self.message('\n\nRun peak detection.')
        super(Mint, self).detect_peaks(**kwargs)

    def set_progress(self, value):
        self.progress_bar.value = value
    
    def export_action(self, b=None, filename=None):
        if filename is None:
            filename = 'MINT__results.xlsx'
            filename = os.path.join(HOME, filename)
        self.export(filename)
        self.message(f'\n\nExported results to: {filename}')




import pandas as pd

from sklearn.metrics import r2_score

from ms_mint.Mint import Mint
from ms_mint.standards import MINT_RESULTS_COLUMNS

mint = Mint(verbose=True)
mint_b = Mint(verbose=True)

class TestClass():
    
    def run_without_files(self):
        mint.run()
    
    def test__add_experimental_data(self):
        assert mint.n_files == 0, mint.n_files
        mint.ms_files = ['tests/data/test.mzXML']
        assert mint.n_files == 1, mint.n_files
        assert mint.files == ['tests/data/test.mzXML']

    def test__add_peaklist(self):
        assert mint.n_peaklist_files == 0, mint.n_peaklist_files
        mint.peaklist_files = 'tests/data/test_peaklist.csv'
        assert mint.peaklist_files == ['tests/data/test_peaklist.csv']
        assert mint.n_peaklist_files == 1, mint.n_peaklist_files
                
    def test__mint_run_standard(self):
        mint.run()
           
    def test__correct_peakAreas(self):
        df_test = pd.read_csv('tests/data/test_peaklist.csv', dtype={'peak_label': str})
        print(mint.results.dtypes)
        df = pd.merge(df_test, mint.results, on='peak_label', suffixes=('_real', '_calc'))
        R2 = r2_score(df.peak_area_real, df.peak_area_calc)
        assert R2 > 0.999, R2
    
    def test__results_is_dataframe(self):
        results = mint.results
        assert isinstance(results, pd.DataFrame), 'Results is not a DataFrame'
    
    def test__results_lenght(self):
        actual = len(mint.results)
        expected = len(mint.peaklist) * len(mint.files)
        assert  expected == actual, f'Length of results ({actual}) does not equal expected length ({expected})'
    
    def test__results_columns(self):
        expected = MINT_RESULTS_COLUMNS
        actual = mint.results.columns
        assert (expected == actual).all(), actual
    
    def test__crosstab_is_dataframe(self):
        ct = mint.crosstab()
        assert isinstance(ct, pd.DataFrame), f'Crosstab is not a DataFrame ({type(ct)}).'

    def test__mint_run_parallel(self):
        mint.files = ['tests/data/test.mzXML']*2
        peaklist = pd.DataFrame({
                      'mz_mean': [100], 
                      'mz_width': [10],
                      'rt_min': [0.1], 
                      'rt_max': [0.2],
                      'intensity_threshold': [0],
                      'peak_label': ['test'],
                      'peaklist_name': ['no-file']})
        mint.peaklist = peaklist
        mint.run(nthreads=2)   
      
    def test__mzxml_equals_mzml(self):
        mint.reset()
        mint.files = ['tests/data/test.mzXML' , 'tests/data/test.mzML']
        mint.peaklist_files = 'tests/data/peaklist_v1.csv'
        mint.run()
        results = []
        for _, grp in mint.results.groupby('ms_file'):
            results.append(grp.peak_area)
        assert (results[0] == results[1]).all(), results

    def test__peaklist_v0_equals_v1(self):
        mint.reset()
        mint.files = ['tests/data/test.mzXML']
        mint.peaklist_files = ['tests/data/peaklist_v0.csv', 
                               'tests/data/peaklist_v1.csv']
        mint.run()
        results = []
        for _, grp in mint.results.groupby('peaklist_name'):
            results.append(grp.peak_area.values)
        assert (results[0] == results[1]).all(), results
    
    def test__status(self):
        mint.status == 'waiting'
    
    def test__run_returns_none_without_peaklist(self):
        mint.reset()
        mint.files = ['tests/data/test.mzXML']
        assert mint.run() is None
    
    def test__run_returns_none_without_ms_files(self):
        mint.reset()
        mint.peaklist_files = 'tests/data/peaklist_v0.csv'
        assert mint.run() is None




import dash_html_components as html
import dash_core_components as dcc


info_txt = '''
Creating chromatograms from mzXML/mzML files can last 
a long time the first time. Try converting your files to 
_feather_ format first.'
'''


pko_layout = html.Div([
    html.H3('Peak Optimization'),
    html.Button('Generate peak previews', id='pko-peak-preview'),
    html.Button('Find closest peaks', id='pko-find-closest-peak'),
    dcc.Markdown('---'),
    html.Div(id='pko-peak-preview-output'),
    dcc.Markdown(id='pko-find-closest-peak-output'),

    html.Div(id='pko-controls'),
    dcc.Dropdown(
        id='pko-dropdown',
        options=[],
        value=None
    ),
    dcc.Loading( dcc.Graph('pko-figure') ),
    dcc.Markdown(id='pko-set-rt-output'),
    html.Button('Set RT to current view', id='pko-set-rt'),

    html.Div([
        html.Button('<< Previous', id='pko-prev'),
        html.Button('Next >>', id='pko-next')],
        style={'text-align': 'center', 'margin': 'auto'})
])


from ms_mint.peak_detection import OpenMSPeakDetection as OMS


def test__peaks_are_close__identical_peaks():
    peak_a = (100, 1, 2, 1.5)
    peak_b = (100, 1, 2, 1.5)
    result = OMS.peaks_are_close(peak_a, peak_b)
    expected = True
    assert result == expected, result


def test__peaks_are_close__different_masses():
    peak_a = (100, 1, 2, 1.5)
    peak_b = (101, 1, 2, 1.5)
    result = OMS.peaks_are_close(peak_a, peak_b)
    expected = False
    assert result == expected, result


def test__peaks_are_close__different_rt():
    peak_a = (100, 1, 2, 1.5)
    peak_b = (100, 2, 3, 2.5)
    result = OMS.peaks_are_close(peak_a, peak_b)
    expected = False
    assert result == expected, result


def test__peaks_are_close__different_rtminmax():
    peak_a = (100, 1, 2, None)
    peak_b = (100, 2, 3, None)
    result = OMS.peaks_are_close(peak_a, peak_b)
    expected = False
    assert result == expected, result


def test__peaks_are_close__close_mass():
    peak_a = (100, 1, 2, None)
    peak_b = (100.001, 1, 2, None)
    result = OMS.peaks_are_close(peak_a, peak_b)
    expected = True
    assert result == expected, result


def test__peaks_are_close__mass_delta():
    peak_a = (100, 1, 2, None)
    peak_b = (100.01, 1, 2, None)
    result = (OMS.peaks_are_close(peak_a, peak_b, max_delta_mz_ppm=10),
              OMS.peaks_are_close(peak_a, peak_b, max_delta_mz_ppm=100))
    expected = (False, True)
    assert result == expected, result


def test__peaks_are_close__rt_delta_rtminmax():
    peak_a = (100, 1, 2, None)
    peak_b = (100.001, 1.1, 2.1, None)
    result = (OMS.peaks_are_close(peak_a, peak_b, max_delta_rt=0.10001),
              OMS.peaks_are_close(peak_a, peak_b, max_delta_rt=0.1))
    expected = (True, False)
    assert result == expected, result


def test__peaks_are_close__rt_delta():
    peak_a = (100, None, None, 1)
    peak_b = (100.001, None, None, 1.1)
    result = (OMS.peaks_are_close(peak_a, peak_b, max_delta_rt=0.10001),
              OMS.peaks_are_close(peak_a, peak_b, max_delta_rt=0.1))
    expected = (True, False)
    assert result == expected, result





import base64
import os
import io

import numpy as np
import pandas as pd

from tqdm import tqdm
from glob import glob

from ms_mint.io import ms_file_to_df
from ms_mint.peaklists import standardize_peaklist, read_peaklists
from ms_mint.io import convert_ms_file_to_feather

from datetime import date



def today():
    return date.today().strftime('%y%m%d')


def parse_ms_files(contents, filename, date, target_dir):
    content_type, content_string = contents.split(',')
    decoded = base64.b64decode(content_string)
    fn_abs = os.path.join(target_dir, filename)
    with open(fn_abs, 'wb') as file:
        file.write(decoded)
    new_fn = convert_ms_file_to_feather(fn_abs)
    print(f'Convert {fn_abs} to {new_fn}')
    if os.path.isfile(new_fn): os.remove(fn_abs)


def parse_pkl_files(contents, filename, date, target_dir):
    content_type, content_string = contents.split(',')
    decoded = base64.b64decode(content_string)
    df = pd.read_csv(io.StringIO(decoded.decode('utf-8')))
    df = standardize_peaklist(df)
    df = df.drop_duplicates()
    return df


def get_dirnames(path):
    dirnames = [ f.name for f in os.scandir(path) if f.is_dir() ]
    return dirnames


def create_chromatograms(ms_files, peaklist, wdir):
    for fn in tqdm(ms_files):
        fn_out = os.path.basename(fn)
        fn_out, _ = os.path.splitext(fn_out) 
        fn_out += '.feather'
        for ndx, row in peaklist.iterrows():
            mz_mean, mz_width = row[['mz_mean', 'mz_width']]
            fn_chro = get_chromatogram_fn(fn, mz_mean, mz_width, wdir)
            if not os.path.isfile(fn_chro):
                create_chromatogram(fn, mz_mean, mz_width, fn_chro)


def create_chromatogram(ms_file, mz_mean, mz_width, fn_out):
    df = ms_file_to_df(ms_file)
    dirname = os.path.dirname(fn_out)
    if not os.path.isdir(dirname): os.makedirs(dirname)
    dmz = mz_mean*1e-6*mz_width
    chrom = df[(df['m/z array']-mz_mean).abs()<=dmz]
    chrom['retentionTime'] = chrom['retentionTime'].round(3)
    chrom = chrom.groupby('retentionTime').max().reset_index()
    chrom[['retentionTime', 'intensity array']].to_feather(fn_out)
    return chrom


def get_chromatogram(ms_file, mz_mean, mz_width, wdir):
    fn = get_chromatogram_fn(ms_file, mz_mean, mz_width, wdir)
    if not os.path.isfile(fn):
        chrom = create_chromatogram(ms_file, mz_mean, mz_width, fn)
    else:
        chrom = pd.read_feather(fn)
    return chrom


def get_chromatogram_fn(ms_file, mz_mean, mz_width, wdir):
    ms_file = os.path.basename(ms_file)
    base, _ = os.path.splitext(ms_file)
    fn = os.path.join(wdir, 'chromato', f'{mz_mean}-{mz_width}'.replace('.', '_'), base, '.feather')
    return fn


def get_peaklist_fn(wdir):
    return os.path.join(wdir, 'peaklist', 'peaklist.csv')


def get_peaklist(wdir):
    return read_peaklists(get_peaklist_fn(wdir)).set_index('peak_label')


def update_peaklist(wdir, peak_label, rt_min=None, rt_max=None):
    peaklist = get_peaklist(wdir)

    if isinstance(peak_label, str):
        if not np.isnan(rt_min):
            peaklist.loc[peak_label, 'rt_min'] = rt_min
        if not np.isnan(rt_max):
            peaklist.loc[peak_label, 'rt_max'] = rt_max

    if isinstance(peak_label, int):
        peaklist = peaklist.reset_index()
        if not np.isnan(rt_min):
            peaklist.loc[peak_label, 'rt_min'] = rt_min
        if not np.isnan(rt_max):
            peaklist.loc[peak_label, 'rt_max'] = rt_max
        peaklist = peaklist.set_index('peak_label')
    peaklist.to_csv(get_peaklist_fn(wdir))


def get_results_fn(wdir):
    return os.path.join(wdir, 'results', 'results.csv')


def get_metadata_fn(wdir):
    fn = os.path.join(wdir, 'metadata','metadata.csv')
    dirname = os.path.dirname(fn)
    if not os.path.isdir(dirname): os.makedirs(dirname)
    return fn

def get_ms_fns(wdir):
    fns = glob(os.path.join(wdir, 'ms_files', '*.feather'))
    return fns


from ms_mint.notebook import Mint

mint = Mint(verbose=True)

class TestClass():
    def test__mint_n_files(self):
        ms_files = ['tests/data/ms_files/fileA.mzXML', 
                    'tests/data/ms_files/fileB.mzxml', 
                    'tests/data/ms_files/fileC.mzML',
                    'tests/data/ms_files/fileD.mzml']
        mint.ms_files = ms_files
        result = mint.n_files
        expect = len(ms_files)
        assert result == expect, f'Expected ({expect}) != result ({result})'



#!/usr/bin/env python

import os
import sys
import subprocess
import multiprocessing
import argparse
from waitress import serve

import ms_mint

from os.path import expanduser
from pathlib import Path as P

HOME = expanduser("~")
DATADIR = str( P(HOME)/'MINT' )


"""
Simple module that monkey patches pkg_resources.get_distribution used by dash
to determine the version of Flask-Compress which is not available with a
flask_compress.__version__ attribute. Known to work with dash==1.16.3 and
PyInstaller==3.6.
"""

from collections import namedtuple

import pkg_resources

IS_FROZEN = hasattr(sys, '_MEIPASS')

# backup true function
_true_get_distribution = pkg_resources.get_distribution
# create small placeholder for the dash call
# _flask_compress_version = parse_version(get_distribution("flask-compress").version)
_Dist = namedtuple('_Dist', ['version'])

def _get_distribution(dist):
    if IS_FROZEN and dist == 'flask-compress':
        return _Dist('1.5.0')
    else:
        return _true_get_distribution(dist)

# monkey patch the function so it can work once frozen and pkg_resources is of
# no help
pkg_resources.get_distribution = _get_distribution



if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='MINT frontend.')

    parser.add_argument('--no-browser', action='store_true', default=False)
    parser.add_argument('--version', default=False, action='store_true')
    parser.add_argument('--data-dir', default=DATADIR)
    parser.add_argument('--debug', default=False, action='store_true')
    parser.add_argument('--port', type=int, default=9999)
    parser.add_argument('--serve-path', default=None, type=str)

    args = parser.parse_args()

    if args.version:
        print('Mint version:', ms_mint.__version__)
        exit()

    url = f'http://localhost:{args.port}'
    
    if not args.no_browser:
        if os.name == 'nt':
            # https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Multiprocessing
            multiprocessing.freeze_support()
            
        if sys.platform in ['win32', 'nt']:
            os.startfile(url)
            
        elif sys.platform=='darwin':
            subprocess.Popen(['open', url])
            
        else:
            try:
                subprocess.Popen(['xdg-open', url])
            except OSError:
                print('Please open a browser on: ', url)

    
    if  args.data_dir is not None: 
        os.environ["MINT_DATA_DIR"] = args.data_dir
    
    if args.serve_path is not None:
        os.environ['MINT_SERVE_PATH'] = args.serve_path

    from ms_mint.app.app import app

    #app.run_server(debug=args.debug, port=args.port)
    serve(app.server, port=9999)


# ms_mint/Mint.py

import os
import numpy as np
import pandas as pd
import time

from pathlib import Path as P

from multiprocessing import Pool, Manager, cpu_count

from .processing import process_ms1_files_in_parallel
from .io import export_to_excel
from ms_mint.standards import MINT_RESULTS_COLUMNS, PEAKLIST_COLUMNS, DEPRECATED_LABELS
from .peaklists import read_peaklists, check_peaklist, standardize_peaklist
from .peak_detection import OpenMSFFMetabo
from .helpers import is_ms_file
from .vis.plotly.plotly_tools import plot_heatmap
from .vis.mpl import plot_peak_shapes, hierarchical_clustering
from .peak_optimization.RetentionTimeOptimizer import RetentionTimeOptimizer
from warnings import simplefilter
from scipy.cluster.hierarchy import ClusterWarning

import ms_mint

class Mint(object):
    def __init__(self, verbose:bool=False, progress_callback=None):
        self._verbose = verbose
        self._version = ms_mint.__version__
        self._progress_callback = progress_callback
        self.reset()
        if self.verbose:
            print('Mint Version:', self.version , '\n')
        self.peak_detector = OpenMSFFMetabo(progress_callback=progress_callback)
        self._rt_optimizer = RetentionTimeOptimizer(self)

    @property
    def verbose(self):
        return self._verbose
    
    @verbose.setter
    def verbose(self, value:bool):
        self._verbose = value
    
    @property
    def version(self):
        return self._version
    
    def reset(self):
        self._files = []
        self._peaklist_files = []
        self._peaklist = pd.DataFrame(columns=PEAKLIST_COLUMNS)
        self._results = pd.DataFrame({i: [] for i in MINT_RESULTS_COLUMNS})
        self._all_df = None
        self._progress = 0
        self.runtime = None
        self._status = 'waiting'
        self._messages = []

    def optimize_retention_times(self, peak_labels=None, **kwargs):
        return self._rt_optimizer.fit_transform(peak_labels=peak_labels, **kwargs)

    def clear_peaklist(self):
        self.peaklist = pd.DataFrame(columns=PEAKLIST_COLUMNS)

    def clear_results(self):
        self.results = pd.DataFrame(columns=MINT_RESULTS_COLUMNS)

    def clear_ms_files(self):
        self.ms_files = []

    def run(self, nthreads=None, rt_margin=.5, mode='standard'):
        '''
        Run MINT with set up peaklist and ms-files.
        ----
        Args
            - nthreads: int or None, default = None
                * None: Run with min(n_cpus, c_files) CPUs
                * 1: Run without multiprocessing on one CPU
                * >1: Run with multiprocessing enabled using nthreads threads.
            - mode: str, default = 'standard'
                * 'standard': calculates peak shaped projected to RT dimension
                * 'express': omits calculation of other features, only peak_areas
        '''
        self._status = 'running'
            
        if (self.n_files == 0) or ( len(self.peaklist) == 0):
            return None

        peaklist = self.peaklist
        if 'rt' in peaklist.columns:
            ndx = ((peaklist.rt_min.isna()) & (~peaklist.rt.isna()))
            peaklist.loc[ndx, 'rt_min'] = peaklist.loc[ndx, 'rt'] - rt_margin
            ndx = ((peaklist.rt_max.isna()) & (~peaklist.rt.isna()))
            peaklist.loc[ndx, 'rt_max'] = peaklist.loc[ndx, 'rt'] + rt_margin
            del ndx

        if nthreads is None:
            nthreads = min(cpu_count(), self.n_files)
            
        if self.verbose: print(f'Run MINT with {nthreads} processes:')
        
        start = time.time()
        if nthreads > 1:
            self.run_parallel(nthreads=nthreads, mode=mode)
        else:
            results = []
            for i, filename in enumerate(self.ms_files):
                args = {'filename': filename,
                        'peaklist': self.peaklist,
                        'q': None, 
                        'mode': mode}
                results.append(process_ms1_files_in_parallel(args))
                self.progress = int(100 * (i / self.n_files))
            self.results = pd.concat(results)
            self.progress = 100
            
        end = time.time()
        self.runtime = ( end - start )
        self.runtime_per_file = (self.runtime / self.n_files)
        self.runtime_per_peak = (self.runtime / self.n_files / len(self.peaklist))
        
        if self.verbose: 
            print(f'Total runtime: {self.runtime:.2f}s')
            print(f'Runtime per file: {self.runtime_per_file:.2f}s')
            print(f'Runtime per peak ({len(self.peaklist)}): {self.runtime_per_peak:.2f}s\n')
            print('Results:', self.results )
        self._status = 'done'

    def detect_peaks(self, **kwargs):
        detected = self.peak_detector.fit_transform(self.ms_files, **kwargs)
        if detected is not None:
            self.peaklist = pd.concat([self.peaklist, detected])

    def run_parallel(self, nthreads=1, mode='standard'):
        pool = Pool(processes=nthreads)
        m = Manager()
        q = m.Queue()
        args = []
        for i, filename in enumerate(self.ms_files):
            args.append({'filename': filename,
                         'peaklist': self.peaklist,
                         'queue': q,
                         'mode': mode})
                   
        results = pool.map_async(process_ms1_files_in_parallel, args)
        
        # monitor progress
        while True:
            if results.ready():
                break
            else:
                size = q.qsize()
                self.progress = 100 * size / self.n_files
                time.sleep(1)
    
        self.progress = 100

        pool.close()
        pool.join()
        results = results.get()
        self.results = pd.concat(results)

    @property
    def messages(self):
        return self._messages
             
    @property
    def status(self):
        return self._status

    @property
    def files(self):
        return self._files

    @property
    def ms_files(self):
        return self._files
    
    @files.setter
    def files(self, list_of_files):
        print('Mint.files is deprecated, please use Mint.ms_files instead!')
        self.ms_files = list_of_files

    @property
    def n_files(self):
        return len(self.ms_files)
    
    @ms_files.setter
    def ms_files(self, list_of_files):
        if isinstance(list_of_files, str):
            list_of_files = [list_of_files]
        list_of_files = [str(P(i)) for i in list_of_files if is_ms_file(i)]
        for f in list_of_files:
            if not os.path.isfile(f): 
                print(f'W File not found ({f})')
        self._files = list_of_files
        if self.verbose:
            print( 'Set files to:\n' + '\n'.join(self.ms_files) + '\n' )

    @property
    def peaklist_files(self):
        return self._peaklist_files

    @peaklist_files.setter
    def peaklist_files(self, list_of_files):
        if isinstance(list_of_files, str):
            list_of_files = [list_of_files]
        if not isinstance(list_of_files, list):
            raise ValueError('Input should be a list of files.')
        for f in list_of_files:
            assert os.path.isfile(f), f'File not found ({f})' 
        self._peaklist_files = list_of_files
        if self.verbose: print( 'Set peaklist files to:\n'.join(self.peaklist_files) + '\n')
        self.peaklist = read_peaklists(list_of_files)

    @property
    def n_peaklist_files(self):
        return len(self.peaklist_files)
    
    @property
    def peaklist(self):
        return self._peaklist

    @peaklist.setter
    def peaklist(self, peaklist):
        peaklist = standardize_peaklist(peaklist)
        errors = check_peaklist(peaklist)
        if len(errors) != 0:
            peaklist = peaklist.drop_duplicates()
            error_string = '\n'.join(errors)
            if self.verbose:
                print(f'Errors in peaklist:\n{error_string}')
            self._messages = errors
        self._peaklist = peaklist
        if self.verbose:
            print('Set peaklists to:\n', self.peaklist.to_string(), '\n')

    @property
    def results(self):
        return self._results
    
    @results.setter
    def results(self, df):
        self._results = df

    @property
    def rt_projections(self):
        return DeprecationWarning('rt_projections is deprecated. Peak shapes are now '
                                 'directly stored in the results table (mint.results).')
        
    def crosstab(self, col_name='peak_area'):
        return pd.crosstab(self.results.ms_file, 
                           self.results.peak_label, 
                           self.results[col_name], 
                           aggfunc=sum).astype(np.float64)
    @property
    def progress_callback(self):
        return self._progress_callback
    
    @progress_callback.setter
    def progress_callback(self, func=None):
        self._progress_callback = func
    
    @property
    def progress(self):
        return self._progress
    
    @progress.setter
    def progress(self, value):
        assert value >= 0, value
        assert value <= 100, value
        self._progress = value
        if self.progress_callback is not None:
            self.progress_callback(value)

    def export(self, filename=None):
        fn = filename
        if fn is None:
            buffer = export_to_excel(self, fn=fn)
            return buffer
        elif fn.endswith('.xlsx'):
            export_to_excel(self, fn=fn)
        elif fn.endswith('.csv'):
            self.results.to_csv(fn, index=False)

    def load(self, fn):
        if self.verbose: print('Loading MINT state')
        if isinstance(fn, str):
            if fn.endswith('xlsx'):
                results = pd.read_excel(fn, sheet_name='Results').rename(columns=DEPRECATED_LABELS)
                ms_files = results.ms_file.drop_duplicates()
                self.results = results
                self.peaklist = pd.read_excel(fn, sheet_name='Peaklist')
                self.ms_files = ms_files

            elif fn.endswith('.csv'):
                results = pd.read_csv(fn).rename(columns=DEPRECATED_LABELS)
                ms_files = results.ms_file.drop_duplicates()
                peaklist = results[
                        [col for col in PEAKLIST_COLUMNS if col in results.columns]
                    ].drop_duplicates()
                self.results = results
                self.ms_files = ms_files
                self.peaklist = peaklist
                return None
        else:
            data = pd.read_csv(fn).rename(columns=DEPRECATED_LABELS)
            if 'ms_file' in data.columns:
                ms_files = data.ms_file.drop_duplicates()
                self.results = data
                self.ms_files = ms_files
            peaklist = data[[col for col in PEAKLIST_COLUMNS if col in data.columns]].drop_duplicates()
            self.peaklist = peaklist

    
    def plot_clustering(self, data=None, title=None, figsize=(8,8), target_var='peak_max',
                        vmin=-3, vmax=3, xmaxticks=None, ymaxticks=None, transpose=False,
                        normalize_by_row=True, transform_func='log1p', 
                        transform_filenames_func='basename', **kwargs):
        '''
        Performs a cluster analysis and plots a heatmap. If no data is provided, 
        data is taken form self.crosstab(target_var).
        First the data is transformed with the transform function (default is log(x+1)).
        If normalize_by_row==True the (transformed) data is also row-normalized (z=(x-μ)/σ)).
        A hierachical cluster analysis is performed with the (transformed) data.
        The original data is then re-ordered accoring to the resulting clusters.
        The result is stored in self.clustered.
        '''
        if len(self.results) == 0:
            return None
        simplefilter("ignore", ClusterWarning)
        if data is None:
            data = self.crosstab(target_var).copy()

        tmp_data = data.copy()

        if transform_func == 'log1p':
            transform_func = np.log1p

        if transform_func is not None:
            tmp_data = tmp_data.apply(transform_func)

        if transform_filenames_func == 'basename':
            transform_filenames_func = os.path.basename

        if transform_filenames_func is not None:
            tmp_data.columns = [transform_filenames_func(i) for i in tmp_data.columns] 

        if normalize_by_row:
            tmp_data = ((tmp_data.T - tmp_data.T.mean()) / tmp_data.T.std())

        if transpose: tmp_data = tmp_data.T    

        clustered, fig, ndx_x, ndx_y = hierarchical_clustering( 
            tmp_data, vmin=vmin, vmax=vmax, figsize=figsize, 
            xmaxticks=xmaxticks, ymaxticks=ymaxticks, **kwargs )

        if not transpose:
            self.clustered = data.iloc[ndx_y, ndx_x]
        else:
            self.clustered = data.iloc[ndx_x, ndx_y]
        return fig
    

    def plot_peak_shapes(self, **kwargs):
        if len(self.results) > 0:
            return plot_peak_shapes(self.results, **kwargs)


    def plot_heatmap(self, target_var='peak_max', **kwargs):
        if len(self.results) > 0:
            return plot_heatmap(self.crosstab(target_var), **kwargs)


import pandas as pd
from matplotlib import pyplot as plt
#from scipy.signal import find_peaks, find_peaks_cwt

def plot_peak_properties(peak_data, show_legend=True, find_peaks=False):
    peak_shape_int = [float(i) for i in peak_data.peak_shape_int.split(',')]
    peak_shape_rt = [float(i) for i in peak_data.peak_shape_rt.split(',')]

    shape = pd.Series(peak_shape_int, index=peak_shape_rt)
    
    if peak_shape_rt is None:
        print('Nothing to plot.')
        return None

    plt.plot(peak_shape_rt, peak_shape_int, color='k', label='Signal')
    
    plt.vlines(peak_data.peak_rt_of_max, 0, 
           peak_data.peak_max, lw=1, colors='k', 
           linestyle='--', label='RT of max', 
           color='grey')
    
    plt.hlines([peak_data.peak_max, peak_data.peak_min], 
            peak_data.rt_min, peak_data.rt_max, 
            color='C0', linewidth=2, label='Min/Max')
    
    plt.hlines([peak_data.peak_median], peak_data.rt_min,
            peak_data.rt_max, colors='C1', label='Median')
    
    plt.hlines([peak_data.peak_mean], peak_data.rt_min, 
            peak_data.rt_max, colors='C2', label='Mean')
    
    plt.hlines([peak_data.peak_int_first, peak_data.peak_int_last], 
            peak_data.rt_min, peak_data.rt_max, colors='y', 
            linestyle='--', linewidth=2, label='Last/first intensity')

    # Find peaks
    if find_peaks:
        n_rolling = 1
        if len(shape) > 25:
            n_rolling = max(1, int(len(shape) / 5))
              
        r_shape = (3*shape.rolling(n_rolling, center=True).max() + 
        shape.rolling(n_rolling, center=True).mean())
       
        a, h = find_peaks(r_shape, height=1e5, rel_height=1e4)
        print(f'Found {len(a)} peaks.')
       
        shape.iloc[a].plot(lw=0, marker='x', ms=5, 
                           mew=1, label='Detected Peaks', color='green')
        
    if show_legend:
        plt.legend(loc=0, bbox_to_anchor=(1,1), fontsize=8)
    
    # Formating the figure
    plt.gca().ticklabel_format(axis='y', style='sci', scilimits=(0,0))
    plt.title(peak_data.peak_label, size=10)
    plt.xlabel('Retention Time [min]')
    plt.ylabel('Intensity')
    if show_legend:
        plt.legend(loc=0, bbox_to_anchor=(1,1), fontsize=8)
    plt.tight_layout()



import os
import pandas as pd
import pytest

from ms_mint.Mint import Mint

mint_a = Mint()
mint_b = Mint()
mint_c = Mint()


@pytest.fixture(scope="session")
def tmp_path(tmpdir_factory):
    path = tmpdir_factory.mktemp("data")
    return path


class _TestExportAndLoad():

    def test__load_example_data(self):
        fn = 'tests/data/demo_results.xlsx'
        mint_a.load(fn)
        assert len(mint_a.results) > 0, 'Could not load example data.'
    
    def test__export_to_excel(self, tmp_path):
        fn = os.path.join(tmp_path, 'results.xlsx')
        mint_a.export(fn)
        assert os.path.isfile(fn), 'Output file not generated'

    def test__export_to_csv(self, tmp_path):
        fn = os.path.join(tmp_path, 'results.csv')
        mint_a.export(fn)
        assert os.path.isfile(fn), 'Output file not generated'

    def test__load_from_excel(self, tmp_path):
        fn = os.path.join(tmp_path, 'results.xlsx')
        print('Filename:', fn)
        mint_b.load(fn)

        actual_results = mint_b.results.round(8)
        expect_results = mint_a.results.round(8)
        actual_files = mint_b.ms_files
        expect_files = mint_a.ms_files

        print('Expected results:\n', expect_results)
        print('Actual results:\n', actual_results)
        print('N diff dtypes:', (actual_results.dtypes != expect_results.dtypes).sum() )
        print('N diff columns:', (actual_results.columns != expect_results.columns).sum() )
        print('N diff index:', (actual_results.columns != expect_results.columns).sum() )
        print('N diff values:', (actual_results.values != expect_results).values.sum() )
        print('N diff values:', (actual_results == expect_results).sum())

        assert all([actual_results.equals(expect_results), 
                    set(actual_files)==set(expect_files)]),\
                'Loaded results differ from original data.'
        mint_b.reset()

    def test__load_from_csv(self, tmp_path):
        fn = os.path.join(tmp_path, 'results.csv')
        print('Filename:', fn)
        mint_c.load(fn)

        actual_results = mint_c.results.round(3)
        expect_results = mint_a.results.round(3)
        actual_files = mint_c.ms_files
        expect_files = mint_a.ms_files

        print(actual_results)
        print(expect_results)

        print(actual_files)
        print(expect_files)

        assert len(actual_results) == len(expect_results), pd.read_csv(fn).to_csv('/home/swacker/test.csv')

        print('Expected results:\n', expect_results)
        print('Actual results:\n', actual_results)
        print('N diff dtypes:', (actual_results.dtypes != expect_results.dtypes).sum() )
        print('N diff columns:', (actual_results.columns != expect_results.columns).sum() )
        print('N diff index:', (actual_results.columns != expect_results.columns).sum() )
        print('N diff values:', (actual_results.values != expect_results).values.sum() )
        print('N diff values:', (actual_results == expect_results).sum())

        ndx = (actual_results.peak_rt_of_max != expect_results.peak_rt_of_max)
        ndx = ndx[ ndx ].index

        print(ndx)
        print(actual_results.loc[ndx])
        print(expect_results.loc[ndx])

        print('Comparison:')
        print('Expected results:\n', expect_results.loc[ndx, 'peak_rt_of_max'])
        print('Actual results:\n', actual_results.loc[ndx, 'peak_rt_of_max'])

        print(expect_results.loc[ndx, 'peak_rt_of_max'] == 
              actual_results.loc[ndx, 'peak_rt_of_max'])
        print((expect_results.loc[ndx, 'peak_rt_of_max'] == 
               actual_results.loc[ndx, 'peak_rt_of_max']).sum() )

        assert all([actual_results.equals(expect_results), 
                    set(actual_files)==set(expect_files)]),\
                'Loaded results differ from original data.'




import io
import os
import re
import base64

import shutil
import tempfile
import subprocess
import platform

from glob import glob
from tqdm import tqdm

import pandas as pd
import numpy as np

import matplotlib
matplotlib.use('Agg')

from matplotlib import pyplot as plt
import seaborn as sns    

import dash
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dash_extensions import Download
from dash_extensions.snippets import send_file
from dash_extensions.enrich import FileSystemCache

from flask_caching import Cache

import plotly.graph_objects as go

from ms_mint.Mint import Mint
from ms_mint.peaklists import read_peaklists
from ms_mint.io import convert_ms_file_to_feather
from ms_mint.vis.plotly.plotly_tools import plot_heatmap
from ms_mint.peak_optimization.RetentionTimeOptimizer import RetentionTimeOptimizer as RTOpt

import ms_mint

from .tools import parse_ms_files, get_dirnames, parse_pkl_files, get_chromatogram,\
    get_metadata_fn, get_results_fn, update_peaklist, today, get_ms_fns, get_peaklist

from .ms_files import ms_layout
from .workspaces import ws_layout
from .peaklist import pkl_layout
from .results import res_layout, res_layout_empty, res_layout_non_empty
from .peak_optimization import pko_layout


config = {
    "DEBUG": True,          # some Flask specific configs
    "CACHE_TYPE": "simple", # Flask-Caching related configs
    "CACHE_DEFAULT_TIMEOUT": 300
}

def get_versions():
    string = ''
    try:
        string += subprocess.getoutput('conda env export --no-build')
    except:
        pass
    return string

ISSUE_TEXT = f'''
%0A%0A%0A%0A%0A%0A%0A%0A%0A
MINT version: {ms_mint.__version__}%0A
OS: {platform.platform()}%0A
Versions:
{get_versions()}
'''




def make_dirs():
    tmpdir = tempfile.gettempdir()
    tmpdir = os.path.join(tmpdir, 'MINT')
    tmpdir = os.getenv('MINT_DATA_DIR', default=tmpdir)
    cachedir = os.path.join(tmpdir, '.cache')
    os.makedirs(tmpdir, exist_ok=True)
    os.makedirs(cachedir, exist_ok=True)
    return tmpdir, cachedir

TMPDIR, CACHEDIR = make_dirs()

fsc = FileSystemCache(CACHEDIR)

app = dash.Dash(__name__, 
    external_stylesheets=[
        dbc.themes.BOOTSTRAP, 
        "https://codepen.io/chriddyp/pen/bWLwgP.css"],
    requests_pathname_prefix=os.getenv('MINT_SERVE_PATH', default='/'),
    routes_pathname_prefix=os.getenv('MINT_SERVE_PATH', default='/')
    )

cache = Cache(app.server, config={
    'CACHE_TYPE': 'filesystem',
    'CACHE_DIR': '/tmp/MINT-cache'
})

app.title = 'MINT'

app.config['suppress_callback_exceptions'] = True

layout = html.Div([
    dcc.Interval(id="progress-interval", n_intervals=0, interval=500, disabled=False),
    html.Button('Run MINT', id='run-mint'),
    html.Button('Download', id='res-download'),
    html.A(href='https://soerendip.github.io/ms-mint/', 
         children=[html.Button('Documentation', id='B_help', style={'float': 'right'})],
         target="_blank"),
    html.A(href=f'https://github.com/soerendip/ms-mint/issues/new?body={ISSUE_TEXT}', 
         children=[html.Button('Issues', id='B_issues', style={'float': 'right'})],
         target="_blank"),   
    Download(id='res-download-data'),
    html.Div(id='run-mint-output'),
    dbc.Progress(id="progress-bar", value=100, style={'margin-bottom': '20px'}),
    html.Div(id='tmpdir', children=TMPDIR, style={'visibility': 'hidden'}),
    html.Div(id='wdir', children=TMPDIR),
    dcc.Tabs(id='tab', value='workspaces', children=[
        dcc.Tab(label='Workspace', value='workspaces'),
        dcc.Tab(label='MS-files', value='msfiles'),
        dcc.Tab(label='Peaklist', value='peaklist'),
        dcc.Tab(label='Peak Optimization', value='pko'),
        #dcc.Tab(label='Quality Control', value='qc'),
        dcc.Tab(label='Heatmap', value='results'),
        #dcc.Tab(label='Cluster Analysis', value='clustering'),
    ]),
    html.Div(id='tab-content', style={'margin': '5%'})
], style={'margin':'2%'})

app.layout = layout

@app.callback(Output('tab-content', 'children'),
              Input('tab', 'value'))
def render_content(tab):
    if tab == 'msfiles':
        return ms_layout
    elif tab == 'workspaces':
        return ws_layout
    elif tab == 'peaklist':
        return pkl_layout
    elif tab == 'results':
        return res_layout
    elif tab == 'pko':
        return pko_layout


@app.callback(
Output('run-mint-output', 'children'),
Input('run-mint', 'n_clicks'),
State('wdir', 'children')
)
def run_mint(n_clicks, wdir):
    if n_clicks is None:
        raise PreventUpdate

    def set_progress(x):
        fsc.set('progress', x)

    mint = Mint(verbose=False, progress_callback=set_progress)
    mint.peaklist_files = os.path.join(wdir, 'peaklist', 'peaklist.csv')
    mint.ms_files = glob( os.path.join(wdir, 'ms_files', '*.*'))
    mint.run()
    mint.export( os.path.join(wdir, 'results', 'results.csv'))


# MS-FILES
@app.callback(
Output('ms-upload-output', 'children'),
Input('ms-upload', 'contents'),
Input('ms-convert-output', 'children'),
State('ms-upload', 'filename'),
State('ms-upload', 'last_modified'),
State('wdir', 'children'))
def ms_upload(list_of_contents, converted, list_of_names, list_of_dates, wdir):
    target_dir = os.path.join(wdir, 'ms_files')
    if list_of_contents is not None:
        n_total = len(list_of_contents)
        n_uploaded = 0

        for i, (c, n, d) in tqdm( enumerate( zip(list_of_contents, list_of_names, list_of_dates) ), total=n_total):
            fsc.set('progress', int( 100*(i+1)/n_total ))
            if n.lower().endswith('mzxml') or n.lower().endswith('mzml'):
                try:
                    parse_ms_files(c, n, d, target_dir)
                    n_uploaded += 1
                except:
                    pass
        return html.P(f'{n_uploaded} files uploaded.')
    

@app.callback(
Output('ms-table', 'data'),
[Input('ms-upload-output', 'children'),
 Input('wdir', 'children'), 
 Input('ms-delete-output', 'children'),
 Input('ms-set-labels', 'n_clicks')],
[State('ms-input', 'value'),
 State('ms-table', 'derived_virtual_selected_rows'),
 State('ms-table', 'derived_virtual_indices')
]
)
def ms_table(value, wdir, files_deleted, set_labels, ms_input, 
    ndxs_selected, ndxs_filtered):   

    target_dir = os.path.join(wdir, 'ms_files')
    ms_files = glob(os.path.join(target_dir, '*.*'), recursive=True)
    data =  pd.DataFrame([{'MS-file': os.path.basename(fn) } for fn in ms_files])

    fn = get_metadata_fn(wdir)
    if os.path.isfile(fn):
        metadata = pd.read_csv(fn)
        data = pd.merge(data, metadata, on='MS-file', how='left')

    prop_id = dash.callback_context.triggered[0]['prop_id']

    if prop_id.startswith('ms-set-labels'):
        if ndxs_selected is []:
            ndxs = ndxs_selected
        else: ndxs = ndxs_filtered
        data.loc[ndxs, 'Label'] = ms_input
    return data.to_dict('records')


@app.callback(
Output('ms-save-output', 'children'),
Input('ms-table', 'data'),
State('wdir', 'children')
)
def ms_save_meta(data, wdir):
    if data is None:
        raise PreventUpdate
    fn = get_metadata_fn(wdir)
    if len(data) > 0:
        pd.DataFrame(data).to_csv(fn, index=False)
        return 'Metadata saved.'
    return ''


@app.callback(
Output('ms-convert-output', 'children'),
Input('ms-convert', 'n_clicks'),
State('wdir', 'children')
)
def ms_convert(n_clicks, wdir):
    target_dir = os.path.join(wdir, 'ms_files')
    if n_clicks is None:
        raise PreventUpdate
    fns = glob(os.path.join(target_dir, '*.*'))
    fns = [fn for fn in fns if not fn.endswith('.feather')]
    n_total = len(fns)
    for i, fn in enumerate( fns ):
        fsc.set('progress', int(100*(i+1)/n_total))
        new_fn = convert_ms_file_to_feather(fn)
        if os.path.isfile(new_fn): os.remove(fn)
    return 'Files converted to feather format.'

@app.callback(
Output('ms-delete-output', 'children'),
Input('ms-delete', 'n_clicks'),
[State('ms-table', 'derived_virtual_selected_rows'),
 State('ms-table', 'data'),
 State('wdir', 'children')]
)
def ms_delete(n_clicks, ndxs, data, wdir):
    if n_clicks is None:
        raise PreventUpdate
    target_dir = os.path.join(wdir, 'ms_files')
    for ndx in ndxs:
        fn = data[ndx]['MS-file']
        fn = os.path.join(target_dir, fn)
        os.remove(fn)
    return 'Files deleted'


# WORKSPACES
@app.callback(
Output('ws-create-output', 'children'),
Input('ws-create', 'n_clicks'),
[State('ws-name', 'value'),
State('tmpdir', 'children')]
)
def create_workspace(n_clicks, ws_name, tmpdir):

    ws_names = get_dirnames(tmpdir)

    if ws_name is None or ws_name == '':
        raise PreventUpdate

    if not re.match('^[\w_-]+$', ws_name):
        return 'Name can only contain: a-z, A-Z, 0-9, -,  _'

    if ws_name not in ws_names:
        os.makedirs(os.path.join(tmpdir, ws_name))
        os.makedirs(os.path.join(tmpdir, ws_name, 'ms_files'))
        os.makedirs(os.path.join(tmpdir, ws_name, 'peaklist'))
        os.makedirs(os.path.join(tmpdir, ws_name, 'results'))
        os.makedirs(os.path.join(tmpdir, ws_name, 'figures'))
        os.makedirs(os.path.join(tmpdir, ws_name, 'chromato'))
        return f'Created workspace "{ws_name}"'

    return 'Nothing'

@app.callback(
Output('ws-table', 'data'),
Input('ws-create-output', 'children'),
Input('tab', 'value'),
Input('ws-delete-output', 'children'),
State('tmpdir', 'children')
)
def ws_table(value, tab, delete, tmpdir):
    ws_names = get_dirnames(tmpdir)
    ws_names =  [{'Workspace': ws_name} for ws_name in ws_names 
                    if not ws_name.startswith('.')]
    return ws_names

@app.callback(
[Output('ws-activate-output', 'children'),
 Output('wdir', 'children')],
Input('ws-activate', 'n_clicks'),
[State('ws-table', 'derived_virtual_selected_rows'),
 State('ws-table', 'data'),
 State('tmpdir', 'children')]
)
def ws_activate(n_clicks, ndx, data, tmpdir):
    fn_ws_info = os.path.join( tmpdir, '.active-workspace')
    if n_clicks is None and os.path.isfile(fn_ws_info):
        with open(fn_ws_info, 'r') as file:
            ws_name = file.read()
    else:
        if ndx is None or len(ndx) != 1:
            raise PreventUpdate
        ndx = ndx[0]
        ws_name = data[ndx]['Workspace']
    
    wdir = os.path.join(tmpdir, ws_name)
    message = f'Workspace __{ws_name}__ activated.'
    
    with open(fn_ws_info, 'w') as file:
        file.write(ws_name)
    
    fsc.set('progress', 0)

    return message, wdir


@app.callback(
 Output('progress-bar', 'value'),
 Input('progress-interval', 'n_intervals'),
)
def set_progress(n):
    return fsc.get('progress')


@app.callback(
Output("ws-delete-popup", "is_open"),
Input("ws-delete", "n_clicks"), 
Input("ws-delete-cancel", "n_clicks"),
Input('ws-delete-confirmed', 'n_clicks'),
State("ws-delete-popup", "is_open")
)
def ws_delete(n1, n2, n3, is_open):
    if n1 is None:
        raise PreventUpdate
    if n1 or n2 or n3:
        return not is_open
    return is_open


@app.callback(
Output('ws-delete-output', 'children'),
Input('ws-delete-confirmed', 'n_clicks'),
[State('ws-table', 'derived_virtual_selected_rows'),
 State('ws-table', 'data'),
 State('tmpdir', 'children')]
)
def ws_delete_confirmed(n_clicks, ndxs, data, tmpdir):
    if n_clicks is None or len(ndxs) == 0:
        raise PreventUpdate
    for ndx in ndxs:
        ws_name = data[ndx]['Workspace']
        dirname = os.path.join(tmpdir, ws_name)
        shutil.rmtree(dirname)
    message = f'Worskpace {ws_name} deleted.'
    return message



# PEAKLIST
@app.callback(
Output('pkl-table', 'data'),
Input('pkl-upload', 'contents'),
[State('pkl-upload', 'filename'),
 State('pkl-upload', 'last_modified'),
 State('wdir', 'children')]
)
def pkl_upload(list_of_contents, list_of_names, list_of_dates, wdir):
    target_dir = os.path.join(wdir, 'peaklist')
    fn = os.path.join( target_dir, 'peaklist.csv')
    if list_of_contents is not None:
        dfs = [parse_pkl_files(c, n, d, target_dir) for c, n, d in
            zip(list_of_contents, list_of_names, list_of_dates) ]
        data = dfs[0].to_dict('records')    
        return data
    elif os.path.isfile(fn):
        return read_peaklists(fn).to_dict('records')   


@app.callback(
Output('pkl-save-output', 'children'),
[Input('pkl-save', 'n_clicks'),
 Input('pkl-table', 'data')],
State('wdir', 'children'))
def plk_save(n_clicks, data, wdir):
    target_dir = os.path.join(wdir, 'peaklist')
    df = pd.DataFrame(data)
    fn = os.path.join( target_dir, 'peaklist.csv')
    df.to_csv(fn)
    return 'Peaklist saved.'



# PEAK OPTIMIZATION
@app.callback(
Output('pko-dropdown', 'options'),
Input('tab', 'value'),
State('wdir', 'children')
)
def pko_controls(tab, wdir):
    if tab != 'pko':
        raise PreventUpdate
    peaklist = get_peaklist( wdir )
    options = [{'label':label, 'value': i} for i, label in enumerate(peaklist.index)]
    return options


@app.callback(
Output('pko-figure', 'figure'),
[Input('pko-dropdown', 'value'),
 Input('pko-set-rt-output', 'children')],
State('wdir', 'children'),
State('pko-figure', 'figure')
)
def pko_figure(peak_label, n_clicks, wdir, fig):
    if peak_label is None:
        raise PreventUpdate
    
    peaklist = get_peaklist( wdir ).reset_index()
    ms_files = get_ms_fns( wdir )
    mz_mean, mz_width, rt, rt_min, rt_max, label = \
        peaklist.loc[peak_label, ['mz_mean', 'mz_width', 'rt', 'rt_min', 'rt_max', 'peak_label']]

    if (rt_min is None) or np.isnan(rt_min): rt_min = rt-0.2
    if (rt_max is None) or np.isnan(rt_max): rt_max = rt+0.2
    if (rt is None) or np.isnan(rt): rt = np.mean([rt_min, rt_max])

    if True or fig is None:
        fig = go.Figure()
        fig.layout.hovermode = 'closest'
        fig.layout.xaxis.range=[rt_min, rt_max]

        fig.update_layout( 
            xaxis=dict( 
                rangeslider=dict( 
                    visible=True
                )
            )
        )
        fig.update_layout(title=label)

    #else:
    #    fig = go.Figure(fig)
    #    fig['data'] = []

    fig.add_vline(rt)
    fig.add_vrect(x0=rt_min, x1=rt_max, line_width=0, fillcolor="green", opacity=0.1)

    n_files = len(ms_files)
    for i, fn in tqdm(enumerate(ms_files)):
        fsc.set('progress', int(100*(i+1)/n_files))
        name = os.path.basename(fn)
        name, _ = os.path.splitext(name)
        chrom = get_chromatogram(fn, mz_mean, mz_width, wdir)
        fig.add_trace(
            go.Scatter(x=chrom['retentionTime'], 
                       y=chrom['intensity array'], 
                       name=name)
        )
        fig.update_layout(showlegend=False)
    return fig


@app.callback(
Output('pko-find-closest-peak-output', 'children'),
Input('pko-find-closest-peak', 'n_clicks'),
State('wdir', 'children'))
def pko_find_closest_peak(n_clicks, wdir):
    if n_clicks is None:
        raise PreventUpdate
    fn_pkl = os.path.join( wdir, 'peaklist', 'peaklist.csv') 
    print('Run peak optimization')
    
    def set_progress(x):
        fsc.set('progress', x)

    mint = Mint(verbose=True, progress_callback=set_progress)
    mint.peaklist_files = fn_pkl
    mint.ms_files = glob( os.path.join( wdir, 'ms_files', '*.*'))
    rtopt = RTOpt(mint)
    new_peaklist = rtopt.fit_transform(margin=1, how='closest')
    new_peaklist.to_csv(fn_pkl)
    print('Done')


@app.callback(
Output('pko-set-rt-output', 'children'),
Input('pko-set-rt', 'n_clicks'),
[State('pko-dropdown', 'value'),
 State('pko-figure', 'figure'),
 State('wdir', 'children')]
)
def pko_set_rt(n_clicks, peak_label, fig, wdir):
    if n_clicks is None:
        raise PreventUpdate
    rt_min, rt_max = fig['layout']['xaxis']['range']
    rt_min, rt_max = np.round(rt_min, 4), np.round(rt_max, 4)
    update_peaklist(wdir, peak_label, rt_min, rt_max)
    return f'Set RT span to ({rt_min},{rt_max})'


@app.callback(
Output('pko-dropdown', 'value'),
Input('pko-prev', 'n_clicks'),
Input('pko-next', 'n_clicks'),
State('pko-dropdown', 'value'),
State('pko-dropdown', 'options')
)
def pko_prev_next(n_prev, n_next, value, options):
    if n_prev is None and n_next is None:
        raise PreventUpdate
    prop_id = dash.callback_context.triggered[0]['prop_id']
    if value is None:
        return 0
    if prop_id.startswith('pko-prev'):
        return (value - 1) % len(options)
    if prop_id.startswith('pko-next'):
        return (value + 1) % len(options)


def fig_to_src():
    out_img = io.BytesIO()   
    plt.savefig(out_img, format='png', bbox_inches='tight')
    plt.close('all')
    out_img.seek(0)  # rewind file
    encoded = base64.b64encode(out_img.read()).decode("ascii").replace("\n", "")
    return "data:image/png;base64,{}".format(encoded)

TIMEOUT = 60
@app.callback(
Output('pko-peak-preview-output', 'children'),
Input('pko-peak-preview', 'n_clicks'),
State('wdir', 'children')
)
@cache.memoize(timeout=TIMEOUT)
def peak_preview(n_clicks, wdir):
    if n_clicks is None:
        raise PreventUpdate
    sns.set_context('paper')
    ms_files = get_ms_fns(wdir)
    peaklist = get_peaklist(wdir)
    n_total = len(peaklist)
    images = []
    for i, (ndx, row) in tqdm( enumerate(peaklist.iterrows()), total=n_total ):
        fsc.set('progress', int(100*(i+1)/n_total))
        mz_mean, mz_width, rt, rt_min, rt_max = row[['mz_mean', 'mz_width', 'rt', 'rt_min', 'rt_max']]

        if np.isnan(rt_min) or rt_min is None: rt_min = rt-0.5
        if np.isnan(rt_max) or rt_max is None: rt_max = rt+0.5

        plt.figure(figsize=(3.5,2.5), dpi=50)
        for fn in ms_files:
            try:        
                fn_chro = get_chromatogram(fn, mz_mean, mz_width, wdir)
                fn_chro = fn_chro[(rt_min < fn_chro['retentionTime']) &
                                  (fn_chro['retentionTime'] < rt_max)]
                plt.plot(fn_chro['retentionTime'], fn_chro['intensity array'], lw=1, color='k')
            except:
                pass
        plt.title(ndx)
        plt.gca().ticklabel_format(axis='y', style='sci', scilimits=(0,0))        
        src = fig_to_src()
        _id = f'peak-preview-{i}'
        images.append(html.Img(id=_id, src=src))
        images.append( dbc.Tooltip(ndx, target=_id, style={'font-size=': 'large'}) )
    images.append( dcc.Markdown('---') )
    return images


# RESULTS
@app.callback(
Output('res-controls', 'children'),
[Input('tab', 'value'),
 Input('res-delete-output', 'children')],
State('wdir', 'children')
)
def res_controls(tab, delete, wdir):
    if tab != 'results':
        raise PreventUpdate
    fn = get_results_fn(wdir)
    if os.path.isfile(fn):
        return res_layout_non_empty
    else: 
        return res_layout_empty

@app.callback(
Output('res-delete-output', 'children'),
Input('res-delete', 'n_clicks'),
State('wdir', 'children')
)
def res_delete(n_clicks, wdir):
    os.remove(get_results_fn(wdir))
    return 'Results file deleted.'



@app.callback(
Output('res-heatmap-figure', 'figure'),
Input('res-heatmap', 'n_clicks'),
[State('res-heatmap-options', 'value'),
 State('wdir', 'children')]
)
def res_heatmap(n_clicks, options, wdir):
    fn = get_results_fn(wdir)
    mint = Mint()
    mint.load(fn)

    data = mint.crosstab('peak_max')

    data.fillna(0, inplace=True)
    fig = plot_heatmap(data, 
        normed_by_cols='normed_by_cols' in options, 
        transposed='transposed' in options, 
        clustered='clustered' in options,
        add_dendrogram='add_dendrogram' in options,
        correlation='correlation' in options, 
        call_show='call_show' in options,
        name='peak_max')

    return fig


@app.callback([
Output('res-download-data', 'data'),
Input('res-download', 'n_clicks'),
State('wdir', 'children')
])
def update_link(n_clicks, wdir):
    if n_clicks is None:
        raise PreventUpdate
    fn = get_results_fn(wdir)
    workspace = os.path.basename( wdir )
    return [send_file(fn, filename=f'{today()}-MINT-results_{workspace}.csv')]


if __name__ == '__main__':
    app.run_server(debug=True, threaded=True, dev_tools_hot_reload_interval=5000,
    dev_tools_hot_reload_max_retry=30)



import pandas as pd
import pyopenms as oms

from numpy import mean, max, min, abs

class OpenMSFFMetabo():
    def __init__(self, progress_callback=None):
        self._feature_map = None
        self._progress = 0
        self._progress_callback = progress_callback

    def fit(self, filenames, max_peaks_per_file=1000):
        try:
            feature_map = oms.FeatureMap()
        except:
            pass
        n_files = len(filenames)
        for i, fn in enumerate(filenames):
            self.progress = 100*(i+1)/n_files
            feature_map += oms_ffmetabo_single_file(
                fn, max_peaks_per_file=max_peaks_per_file
            )
        self._feature_map = feature_map
    
    @property
    def progress(self):
        return self._progress
    
    @progress.setter    
    def progress(self, x):
        self._progress = x
        if self._progress_callback is not None:
            self._progress_callback(x)

    def transform(self, min_quality=1e-3, condensed=True, 
                  max_delta_mz_ppm=10, max_delta_rt=0.1):

        features = []
        n_total = self._feature_map.size()
        for i, feat in enumerate(self._feature_map):    

            self.progress = 100*(i+1)/n_total

            quality = feat.getOverallQuality()

            if ( min_quality is not None ) and ( quality < min_quality ):
                continue

            mz = feat.getMZ()
            rt = feat.getRT() / 60
            rt_width = feat.getWidth() / 60
            rt_min = max([0, (rt - rt_width)])
            rt_max = (rt + rt_width)
            
            data = {'peak_label': f'mz:{mz:07.4f}-rt:{rt:03.1f}',
                    'mz_mean': mz, 'mz_width': 10,
                    'rt_min': rt_min, 'rt_max': rt_max, 'rt': rt, 
                    'intensity_threshold': 0,
                    'oms_quality': quality,
                    'peaklist': 'OpenMSFFMetabo'}
            
            features.append(data)

        peaklist = pd.DataFrame(features)
        peaklist = peaklist.reset_index(drop=True)

        if condensed:
            peaklist = condense_peaklist(peaklist, 
                        max_delta_mz_ppm=max_delta_mz_ppm, 
                        max_delta_rt=max_delta_rt,
                        progress_callback=self._progress_callback)  
        return peaklist


    def fit_transform(self, filenames, max_peaks_per_file=1000, 
                      min_quality=1e-3, condensed=True, 
                      max_delta_mz_ppm=10, max_delta_rt=0.1, 
                      progress_callback=None):

        self.fit(filenames, max_peaks_per_file=max_peaks_per_file)
        return self.transform(min_quality=1e-3, condensed=True, 
                              max_delta_mz_ppm=10, max_delta_rt=0.1)


def oms_ffmetabo_single_file(filename, max_peaks_per_file=5000):

    feature_map = oms.FeatureMap()
    mass_traces = []
    mass_traces_split = []
    mass_traces_filtered = []
    exp = oms.MSExperiment()
    peak_map = oms.PeakMap()
    options = oms.PeakFileOptions()
    
    options.setMSLevels([1])

    if filename.lower().endswith('.mzxml'):
        fh = oms.MzXMLFile()

    elif filename.lower().endswith('.mzml'):
        fh = oms.MzMLFile()
    else:
        assert False, filename

    fh.setOptions(options)

    # Peak map
    fh.load(filename, exp)

    #for chrom in exp.getChromatograms():
    #    peak_map.addChrom(chrom)

    for spec in exp.getSpectra():
        peak_map.addSpectrum(spec)

    mass_trace_detect = oms.MassTraceDetection()
    mass_trace_detect.run(peak_map, mass_traces, max_peaks_per_file)

    elution_peak_detection = oms.ElutionPeakDetection()
    elution_peak_detection.detectPeaks(mass_traces, mass_traces_split)

    feature_finding_metabo = oms.FeatureFindingMetabo()
    feature_finding_metabo.run(
                mass_traces_split,
                feature_map,
                mass_traces_filtered)

    feature_map.sortByOverallQuality()
    return feature_map


def condense_peaklist(peaklist, max_delta_mz_ppm=10, max_delta_rt=0.1, progress_callback=None):
    cols = ['mz_mean', 'rt_min', 'rt_max', 'rt']
    peaklist = peaklist.sort_values(cols)[cols]

    n_before = len(peaklist)
    n_after = None

    while n_before != n_after:
        n_before = len(peaklist)
        new_peaklist = pd.DataFrame(columns=cols)

        for i, (ndx_a, peak_a) in enumerate( peaklist.iterrows() ):
            if progress_callback is not None:
                progress_callback(100*(i+1)/n_before)
            mz_a, rt_min_a, rt_max_a, rt_a = peak_a
            merged = False
            for ndx_b, peak_b in new_peaklist[ abs(new_peaklist.mz_mean - mz_a) < 0.01 ].iterrows():
                if ( peaks_are_close(peak_a, peak_b, 
                                    max_delta_mz_ppm=max_delta_mz_ppm, 
                                    max_delta_rt=max_delta_rt) ):
                    
                    merged_peak = merge_peaks(peak_a, peak_b)
                    new_peaklist.loc[ndx_b, cols] = merged_peak
                    merged = True
                    break        
            if not merged:
                new_peaklist = pd.concat([new_peaklist, as_df(peak_a)])\
                                .sort_values(cols).reset_index(drop=True)
        peaklist = new_peaklist
        n_after = len(new_peaklist)
        
    return fix_peaklist(new_peaklist.reset_index(drop=True))


def merge_peaks(peak_a, peak_b):
    mz_a, rt_min_a, rt_max_a, rt_a = peak_a
    mz_b, rt_min_b, rt_max_b, rt_b = peak_b
    return (mean([mz_a, mz_b]), 
            min([rt_min_a, rt_min_b, rt_max_a, rt_max_b]), 
            max([rt_min_a, rt_min_b, rt_max_a, rt_max_b]), 
            mean([rt_a, rt_b]))


def peaks_are_close(peak_a, peak_b, max_delta_mz_ppm=10, max_delta_rt=0.1):
    mz_a, rt_min_a, rt_max_a, rt_a = peak_a
    mz_b, rt_min_b, rt_max_b, rt_b = peak_b
    mz_limit = max([mz_a*max_delta_mz_ppm*1e-6, 
                    mz_b*max_delta_mz_ppm*1e-6])    
    if rt_a is None or rt_b is None:
        rt_mean_a = mean([rt_min_a, rt_max_a])
        rt_mean_b = mean([rt_min_b, rt_max_b])
        rt_delta = abs(rt_mean_a - rt_mean_b)
    else:
        rt_delta = abs(rt_a-rt_b)   
    if abs(mz_a - mz_b) < mz_limit:
        if rt_delta < max_delta_rt:
            return True
    return False


def as_df(peak):
    mz, rt_min, rt_max, rt = peak
    data = dict(mz_mean=mz, rt_min=rt_min, rt_max=rt_max, rt=rt)
    df = pd.DataFrame( data, index=[0] )
    return df


def fix_peaklist(peaklist):
    if 'peak_label' not in peaklist.columns:     
        peaklist['peak_label'] = ( peaklist.index.astype(str) + '_' +
                                   'mz:' + peaklist.mz_mean.apply(lambda x: f'{x:7.3f}') + '_' +
                                   'rt:' + peaklist['rt'].apply(lambda x: f'{x:3.1f}') )
    if 'mz_width' not in peaklist.columns:     
        peaklist['mz_width'] = 10
    if 'peaklist_name' not in peaklist.columns:
        peaklist['peaklist_name'] = 'OMS'
    if 'intensity_threshold' not in peaklist.columns:
        peaklist['intensity_threshold'] = 0
    if 'rt' not in peaklist.columns:
        peaklist['rt'] = peaklist[['rt_min', 'rt_max']].mean(axis=1)
    if 'rt_span' not in peaklist.columns:
        peaklist['rt_span'] = peaklist.rt_max - peaklist.rt_min
    return peaklist



import dash_html_components as html
import dash_core_components as dcc
from dash_table import DataTable
from ms_mint.standards import PEAKLIST_COLUMNS

columns = [{"name": i, "id": i, 
            "selectable": True}  for i in PEAKLIST_COLUMNS] 


pkl_table = DataTable(
                id='pkl-table',
                columns=columns,
                data=None,
                sort_action="native",
                sort_mode="single",
                row_selectable=False,
                row_deletable=True,
                editable=True,
                column_selectable=False,
                selected_rows=[],
                page_action="native",
                page_current= 0,
                page_size= 30,
                filter_action='native',                
                style_table={'overflowX': 'scroll'},
                style_as_list_view=True,
                style_cell={'padding-left': '5px', 
                            'padding-right': '5px'},
                style_header={'backgroundColor': 'white',
                              'fontWeight': 'bold'},
                export_format='csv',
                export_headers='display',
                merge_duplicate_headers=True
            ) 


pkl_layout = html.Div([
    html.H3('Peaklist'),
    dcc.Upload(
            id='pkl-upload',
            children=html.Div([
                'Drag and Drop or ',
                html.A('Select Files')
            ]),
            style={
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin': '10px'
            },
            # Allow multiple files to be uploaded
            multiple=True
        ),
    html.Div(id='pkl-upload-output'),
    html.Div(id='pkl-save-output'),

    pkl_table, 
    html.Button('Save peaklist', id='pkl-save'),
    html.Button('Delete selected peaks', id='pkl-delete', style={'float': 'right'}),
])



import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from scipy.signal import find_peaks, argrelmin, peak_widths

class RetentionTimeOptimizer():
    def __init__(self, mint):
        self._mint = mint
        self._interim_peaklist = self._mint.peaklist.copy()
        self._ms_files = mint.ms_files
        
    def init_rt_minmax(self, peaklist, margin=None):
        '''
        '''
        peaklist = peaklist.copy()
        cols = peaklist.columns
        for ndx, row in peaklist.iterrows():
            if margin is None:
                margin = 0.2
            if (row['rt_min'] is not None) and (row['rt_max'] is not None):
                peaklist.loc[ndx, 'rt_min'] = max( 0, row['rt_min'] - margin)
                peaklist.loc[ndx, 'rt_max'] = row['rt_max'] + margin
            else: 
                assert 'rt' in cols, 'Missing column for retention time (rt)'
                if margin is None:
                    margin = 0
                peaklist.loc[ndx, 'rt_min'] = max( 0, row['rt'] - margin)
                peaklist.loc[ndx, 'rt_max'] = row['rt'] + margin
            # I f rt is none set it to the mean of rt_min and rt_max
            peaklist.loc[peaklist['rt'].isna(), 'rt'] = \
                peaklist.loc[peaklist['rt'].isna(), ['rt_min', 'rt_max']].mean(axis=1)
        self._interim_peaklist = peaklist
        return peaklist

    def fit_transform(self, margin=None, **kwargs):
        self.init_rt_minmax(self._mint.peaklist, margin=margin)
        self._mint.peaklist = self._interim_peaklist
        self._mint.run()
        peaklist, figures = self.fit(**kwargs)
        self._mint.peaklist = peaklist
        self._mint.clear_results()
        return peaklist
        
    def fit(self, show_plots=False, **kwargs):
        return optimize_retention_times(self._mint.results, self._interim_peaklist, **kwargs)

    def get_peaklist(self):
        return self._mint.peaklist


def optimize_retention_times(results, peaklist, create_plots=False, show_plots=True, how='closest', **kwargs):
    print('Optimize RT')

    assert how in ['closest', 'max']

    interim_peaklist = peaklist.set_index('peak_label').copy()
    figures = {}

    for label, grp in results.groupby('peak_label'):
        rt = grp['rt'].values[0]
        rt_min = grp['rt_min'].values[0]
        rt_max = grp['rt_max'].values[0]
        
        grp = grp[grp.peak_n_datapoints > 20]
        
        #if len(grp) < 20:
        #    print('Group to small')
        #    continue

        df = pd.DataFrame({'rt': grp['peak_shape_rt'].apply(lambda x: x.split(',')).explode('peak_shape_rt').astype('float'),
                           'intensity': grp['peak_shape_int'].apply(lambda x: x.split(',')).explode('peak_shape_int').astype('int')})
        if len(df)<20:
            print('Smoothed to small')
            continue

        smoothed = df.groupby('rt').mean().rolling(8, center=True).mean().dropna().reset_index()

        t_min = smoothed.rt.min()
        t_max = smoothed.rt.max()
        t_span = t_max - t_min
        
        dt = 0.01
        interpolated = smoothed.set_index('rt')\
                .reindex(np.arange(0, t_max, dt))\
                .interpolate().rolling(4, center=True)\
                .mean().reset_index().dropna() 

        interpolated = interpolated[interpolated.rt >= t_min]
        interpolated = interpolated[interpolated.rt <= t_max]
        
        t_min = interpolated.rt.min()
        t_max = interpolated.rt.max()
        t_span = t_max - t_min
        
        if (t_min is np.NaN) or (t_max is np.NaN):
            print('Times are NaN', t_min, t_max)
            continue
        
        x_to_t = lambda x: ((t_span * np.array(x) / len(interpolated)) + t_min).flatten()
        
        assert np.abs(x_to_t(0) - t_min) <= dt, (x_to_t(0), t_min)
        assert np.abs(x_to_t(len(interpolated)) - t_max) <= dt, (x_to_t(len(interpolated)), t_max)
        
        t = interpolated['rt']
        x = interpolated['intensity'].values
        
        ndx_peaks, _ = find_peaks(x, prominence=(1000, None))
        
        if len(ndx_peaks) == 0:
            print('No peaks found.')
            continue
        t_peaks = t.iloc[ndx_peaks].values
        x_peaks = x[ndx_peaks]
        
        ndx_minima = argrelmin(x)
        t_minima = x_to_t(ndx_minima)
        x_minima = x[ndx_minima]
        
        results_half  = peak_widths(x, ndx_peaks, rel_height=.5)
        results_bottom  = peak_widths(x, ndx_peaks, rel_height=0.98)
        

        if how == 'closest':
            ndx_selected_peak = (np.argmin(np.abs(t_peaks - rt)))
        elif how == 'max':
            ndx_selected_peak = (np.argmax(x_peaks))


        t_0 = x_to_t(results_bottom[2][ndx_selected_peak])
        t_1 = x_to_t(results_bottom[3][ndx_selected_peak])
        
        t_selected_peak = t_peaks[ndx_selected_peak]
        x_selected_peak = x_peaks[ndx_selected_peak]
        
        for t_m in t_minima:
            if (t_0 < t_m) & (t_m < t_selected_peak):
                t_0 = t_m
            if (t_m > t_selected_peak) & (t_m < t_1):
                t_1 = t_m
                
        t_buffer = 0.01
        t_0 -= t_buffer
        t_1 += t_buffer
        
        interim_peaklist.loc[label, 'rt_min'] = np.round(t_0, 3)
        interim_peaklist.loc[label, 'rt_max'] = np.round(t_1, 3)

        if create_plots:       
            fig = plt.figure()
            plt.plot(smoothed.rt, smoothed.intensity, label='Smoothed')
            plt.plot(interpolated.rt, interpolated.intensity, label='Interpolated')
            plt.plot(t_peaks, x_peaks, "x")

            plt.plot(t_minima, x_minima, "|", ms=10, label='Detected minima')

            plt.hlines(results_half[1], x_to_t(results_half[2]), x_to_t(results_half[3]), color="C2")
            plt.hlines(results_bottom[1], x_to_t(results_bottom[2]), x_to_t(results_bottom[3]), color="C2")
            
            plt.vlines([rt_min, rt_max], 0, max(x), label='RT span')
            plt.vlines([t_min, t_max], 0, max(x), ls='--', label='RT span smoothed')
            
            plt.vlines([rt], 0, max(x), color='C4', lw=2, label='Original RT')

            plt.plot(t_selected_peak, x_selected_peak, "o", color='red', ms=10, label='Selected peak')
            
            plt.hlines(0, t_0, t_1, lw=2, label='Selected RT span')

            plt.title(label)
            plt.xlabel('Retention Time [min]')
            plt.ylabel('Peak Intensity')
            plt.legend()
            plt.ticklabel_format(axis='y', style='sci', scilimits=(0,0))
            plt.grid()
            figures['label'] = fig
            
            if show_plots: plt.show()

    return interim_peaklist.reset_index(), figures


import dash_html_components as html
import dash_core_components as dcc


res_layout = html.Div([
    dcc.Markdown(id='res-delete-output'),
    html.Div(id='res-upload-output'),
    html.Div(id='res-controls'),
    html.Div(id='res-output')
])


heatmap_options = [
    { 'label': 'Normalized by biomarker', 'value': 'normed_by_cols'},
    { 'label': 'Cluster', 'value': 'clustered'},
    { 'label': 'Dendrogram', 'value': 'add_dendrogram'},
    { 'label': 'Transposed', 'value': 'transposed'},
    { 'label': 'Correlation', 'value': 'correlation'},
    { 'label': 'Show in new tab', 'value': 'call_show'}]


res_layout_empty = html.Div([
    dcc.Upload(
            id='res-upload',
            children=html.Div([
                'Drag and Drop or ',
                html.A('Select Files')
            ]),
            style={
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin': '10px'
            },
            # Allow multiple files to be uploaded
            multiple=False
        ),
])

res_layout_non_empty = html.Div([
    html.Button('Delete results', id='ms-delete', style={'float': 'right'}),
    html.H3('Heatmap'),
    html.Button('Heatmap', id='res-heatmap'),
    dcc.Dropdown(id='res-heatmap-options', value=[],
        options=heatmap_options, multi=True),
    dcc.Loading( 
        dcc.Graph(id='res-heatmap-figure', 
                  style={'margin-top': '50px'}) ),
])


import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc

from dash_table import DataTable


ws_table = html.Div(id='ws-table-container', 
    style={'min-height':  100, 'margin': '5%'},
    children=[
        DataTable(id='ws-table',
            columns=[ {"name": i, "id": i, "selectable": True}  
                                for i in ['Workspace']],
                    data=[],
                    row_selectable='single',
                    row_deletable=False,
                    style_cell={'textAlign': 'left'},
                    sort_action='native'
                    )
])


ws_layout = html.Div([
    html.H3('Workspaces'),
    dcc.Input(id='ws-name', placeholder='New workspace name'),
    html.Button('Create Workspace', id='ws-create'),
    html.P(id='ws-create-output'),
    dcc.Markdown(id='ws-activate-output'),
    html.P(id='ws-delete-output'),
    ws_table,
    html.Button('Activate', id='ws-activate'),
    html.Button('Delete', id='ws-delete', style={'float':'right'}),
    dbc.Modal(
            [
                dbc.ModalHeader("Delete Workspace"),
                dbc.ModalBody("This will delete all files and results in the selected workspace."),
                dbc.ModalFooter(
                    html.Div([
                        dbc.Button("Delete", id="ws-delete-confirmed", style={'margin-right': '10px'}),
                        dbc.Button("Cancel", id="ws-delete-cancel")
                    ])
                ),
            ],
            id="ws-delete-popup",
        ),
])




from setuptools import setup, find_packages

import versioneer


with open("README.md", "r") as fh:
    long_description = fh.read()


install_requires = [
    'pandas>=1',
    'plotly',
    'lxml',
    'matplotlib',
    'pandoc',
    'plotly',
    'plotly_express',
    'dash',
    'scipy==1.4',
    'setuptools',
    'statsmodels',
    'flask',
    'pyteomics',
    'openpyxl',
    'colorlover',
    'dash-bootstrap-components',
    'dash-extensions',
    'scikit-learn',
    'xlrd',
    'ipywidgets',
    'pyopenms',
    'pymzml',
    'tqdm',
    'seaborn',
    'ipyfilechooser',
    'waitress',
    'pyarrow',
    'flask-compress'] 

config = {
    'name': 'ms-mint',
    'version': versioneer.get_version(),
    'cmdclass': versioneer.get_cmdclass(),
    'description': 'Metabolomics Integrator (Mint)',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'author': 'Soren Wacker',
    'url': 'https://github.com/soerendip/ms-mint',
    'author_email': 'swacker@ucalgary.ca',
    'scripts': ['scripts/Mint.py'],
    'packages': find_packages(),
    'data_files': [('scripts', ['scripts/Mint.py']),
                   ('static', ['static/Standard_Peaklist.csv'])],
    'classifiers': [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
   'python_requires': '>=3.5',
   'install_requires': install_requires,
}

setup(**config)



# ms_mint/peaklists.py

import os
import pandas as pd
import numpy as np

from .standards import PEAKLIST_COLUMNS, DEPRECATED_LABELS
from .helpers import dataframe_difference


def read_peaklists(filenames):
    '''
    Extracts peak data from csv files that contain peak definitions.
    CSV files must contain columns: 
        - 'peak_label': str, unique identifier
        - 'peakMz': float, center of mass to be extracted in [Da]
        - 'peakMzWidth[ppm]': float, with of mass window in [ppm]
        - 'rtmin': float, minimum retention time in [min]
        - 'rtmax': float, maximum retention time in [min]
    -----
    Args:
        - filenames: str or PosixPath or list of such with path to csv-file(s)
    Returns:
        pandas.DataFrame in peaklist format
    '''
    if isinstance(filenames, str):
        filenames = [filenames]

    peaklist = []
    for fn in filenames:
        if fn.endswith('.csv'):
            df = pd.read_csv(fn)
        elif fn.endswith('.xlsx'):
            df = pd.read_excel(fn)
        df['peaklist_name'] = os.path.basename(fn)
        df = standardize_peaklist(df)
        peaklist.append(df)
    peaklist = pd.concat(peaklist)
    return peaklist


def standardize_peaklist(peaklist):
    peaklist = peaklist.rename(columns=DEPRECATED_LABELS)
    cols = peaklist.columns
    if 'intensity_threshold' not in cols:
        peaklist['intensity_threshold'] = 0
    if 'mz_width' not in cols:
        peaklist['mz_width'] = 10
    if 'peaklist_name' not in cols:
        peaklist['peaklist_name'] = 'unknown'
    for c in ['rt', 'rt_min', 'rt_max']:
        if c not in cols:
            peaklist[c] = None
    del c
    
    if 'peak_label' not in cols:
        peaklist['peak_label'] = [f'C_{i}' for i in range(len(peaklist)) ]        

    peaklist['peak_label'] = peaklist['peak_label'].astype(str)
    peaklist.index = range(len(peaklist))
    return peaklist[PEAKLIST_COLUMNS]

def update_retention_time_columns(peaklist):
    for ndx, row in peaklist.iterrows():
        if row['rt'] is not None:
            if ['rt_min'] is None:
                peaklist.loc[ndx, 'rt_min'] = 5 #max( 0, row['rt'] - 0.2 )
            if row['rt_max'] is None:
                peaklist.loc[ndx, 'rt_max'] = row['rt'] + 0.2
        else:
            if (row['rt_min'] is not None) & (row['rt_max'] is not None):
                peaklist.loc[ndx, 'row'] = row[['rt_min', 'rt_max']].mean(axis=1)



def check_peaklist(peaklist):
    '''
    Test if 
    1) peaklist has right type, 
    2) all columns are present and 
    3) dtype of column peak_label is string
    Returns a list of strings indicating identified errors.
    If list is empty peaklist is OK.
    '''
    errors = []
    if not isinstance(peaklist, pd.DataFrame):
        errors.append('Peaklist is not a dataframe.')
    peaklist[PEAKLIST_COLUMNS]
    if not (peaklist.dtypes['peak_label'] == np.dtype('O')):
        errors.append('Provided peak labels are not strings.', 
                       peaklist.dtypes['peak_label'] )
    if not peaklist.peak_label.value_counts().max() == 1:
        errors.append('Provided peak labels are not unique.')
    return errors


def generate_grid_peaklist(masses, dt, rt_max=10, 
                           mz_ppm=10, intensity_threshold=0):
    '''
    Creates a peaklist from a list of masses.
    -----
    Args:
        - masses: iterable of float values
        - dt: float or int, size of peak windows in time dimension [min]
        - rt_max: float, maximum time [min]
        - mz_ppm: width of peak window in m/z dimension
            mass +/- (mz_ppm * mass * 1e-6)
    '''
    rt_cuts = np.arange(0, rt_max+dt, dt)
    peaklist = pd.DataFrame(index=rt_cuts, columns=masses).unstack().reset_index()
    del peaklist[0]
    peaklist.columns = ['mz_mean', 'rt_min']
    peaklist['rt_max'] = peaklist.rt_min+(1*dt)
    peaklist['peak_label'] =  peaklist.mz_mean.apply(lambda x: '{:.3f}'.format(x))\
                              + '__' + peaklist.rt_min.apply(lambda x: '{:2.2f}'.format(x))
    peaklist['mz_width'] = mz_ppm
    peaklist['intensity_threshold'] = intensity_threshold
    peaklist['peaklist_name'] = 'Generated'
    return peaklist


def diff_peaklist(old_pklist, new_pklist):
    df = dataframe_difference(old_pklist, new_pklist)
    df = df[df['_merge'] == 'right_only']
    return df.drop('_merge', axis=1)


import ipywidgets as widgets
from ipywidgets import HTML
import traitlets
from IPython.display import display
from tkinter import Tk, filedialog

class SelectFilesButton(widgets.Button):
    """A file widget that leverages tkinter.filedialog."""

    def __init__(self, text='Button', default_color=None, callback=None, assign_to=None):
        super(SelectFilesButton, self).__init__()
        # Add the selected_files trait
        self.add_traits(files=traitlets.traitlets.List())
        # Create the button.
        self.description = text
        self.icon = "square-o"
        self.style.button_color = default_color
        # Set on click behavior.
        self.on_click(self.do_stuff)
        self.callback = callback
        self.assign_to = assign_to
        display(HTML("<style>textarea, input { font-family: monospace; }</style>"))
        display(HTML("<style>.container { width:%d%% !important; }</style>" %90))
   
    def do_stuff(self, b):
        try:
            self.select_files(b)
        except:
            return None
        if self.assign_to is not None:
            self.assign_to = b.files
        if self.callback is not None:
            self.callback()

    @staticmethod
    def select_files(b):
        """Generate instance of tkinter.filedialog.

        Parameters
        ----------
        b : obj:
            An instance of ipywidgets.widgets.Button 
        """
        root = Tk()
        root.withdraw()
        root.call('wm', 'attributes', '.', '-topmost', True)
        b.files = filedialog.askopenfilename(multiple=True)



import pandas as pd
from ms_mint import peaklists
from ms_mint.standards import PEAKLIST_COLUMNS

def test__read_peaklists():
    pass


def test__standardize_peaklist():
    peaklist = pd.DataFrame({
        'peakLabel': ['A'],
        'peakMz': [100],
        'rtmin': [1],
        'rtmax': [2],
        'peaklist': ['TEST']
    })
    
    expected = pd.DataFrame(
        {'peak_label': {0: 'A'},
         'mz_mean': {0: 100},
         'mz_width': {0: 10},
         'rt_min': {0: 1},
         'rt_max': {0: 2},
         'rt': {0: None},
         'intensity_threshold': {0: 0},
         'peaklist_name': {0: 'TEST'}}, index=range(0,1)
    )[PEAKLIST_COLUMNS]

    result = peaklists.standardize_peaklist(peaklist)
    print(result.values)
    print(expected.values)    
    print(result.index)
    print(expected.index)
    assert result.equals(expected)

def test__check_peaklist():
    pass


def test__peak_window_from_peaklist():
    pass


def test__generate_grid_peaklist():
    peaklist = peaklists.generate_grid_peaklist(
        [115], .1, intensity_threshold=1000
    )
    assert peaklist is not None



import io

import numpy as np
import pandas as pd
import uuid

from datetime import date, datetime
from flask import send_file
from functools import lru_cache
from glob import glob
from tkinter import Tk, filedialog
from pathlib import Path as P

from os.path import basename, isfile, abspath, join

import dash
import dash_bootstrap_components as dbc

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dash_table import DataTable

from ..Mint import Mint
from .Layout import Layout
from .button_style import button_style

from ms_mint.vis.plotly.plotly_tools import plot_peak_shapes, plot_peak_shapes_3d, plot_heatmap
   
from ms_mint.peaklists import read_peaklists, standardize_peaklist, diff_peaklist
from ms_mint.standards import PEAKLIST_COLUMNS
from ms_mint.helpers import remove_all_zero_columns, sort_columns_by_median

mint = Mint()

app = dash.Dash(__name__, external_stylesheets=[
    dbc.themes.BOOTSTRAP, 
    "https://codepen.io/chriddyp/pen/bWLwgP.css"
                                                ])
app.title = 'MINT'
app.layout = Layout

app.config['suppress_callback_exceptions'] = True

@app.callback(
    [Output("progress-bar", "value"), 
     Output("progress", 'children')],
    [Input("progress-interval", "n_intervals")])
def update_progress(n):
    progress = mint.progress
    # only add text after 5% progress to ensure text isn't squashed too much
    return progress, (f"Progress: {progress} %" if progress >= 5 else "")


### Load MS-files
@app.callback(
    [Output('B_add_files', 'value'),
     Output('table-ms-files', 'data')],
    [Input('B_add_files', 'n_clicks')],
    [State('files-check', 'value'),
     State('table-ms-files', 'data')] )
def select_files(n_clicks, options, ms_files):
    files = []
    if n_clicks is not None:
        root = Tk()
        root.withdraw()
        root.call('wm', 'attributes', '.', '-topmost', True)
        if 'by-dir' not in options:
            files = filedialog.askopenfilename(multiple=True)
            files = [abspath(i) for i in files]
            for i in files:
                assert isfile(i)
        else:
            dir_ = filedialog.askdirectory()
            if isinstance(dir_ , tuple):
                dir_ = []
            if len(dir_) != 0:
                files = glob(join(dir_, join('**', '*.mz*ML')), recursive=True)
            else:
                files = []
        if len(files) != 0:
            mint.ms_files += files
            #mint.progress = 0
        root.destroy()
        ms_files +=  [{'MS-files': fn } for fn in files]
    ms_files = [i for n, i in enumerate(ms_files) if i not in ms_files[n + 1:]]
    return str(n_clicks), ms_files

### Clear files
@app.callback(
    Output('B_reset', 'value'),
    [Input('B_reset', 'n_clicks')])
def clear_files(n_clicks):
    if n_clicks is None:
        raise PreventUpdate
    mint.reset()
    return str(n_clicks)
    
### Update n-files text when files added or cleared
@app.callback(
    [Output('files-text', 'children'),
     Output('n_files_selected', 'children')],
    [Input('B_add_files', 'value'),
     Input('B_reset', 'value'),
     Input('table-ms-files', 'data')])    
def update_files_text(n_clicks, n_clicks_clear, ms_files):
    return '{} data files selected.'.format(mint.n_files), len(ms_files)


### Edit peaklist
@app.callback(
    [Output('table-peaklist-container', 'children'),
     Output('peaklist-div', 'style')],
    [Input('B_peaklists', 'n_clicks'),
     Input('B_reset', 'value'),
     Input('B_add_peak', 'n_clicks'),
     Input('B_detect_peaks', 'n_clicks'),
     Input('B_clear_peaklist', 'n_clicks'),
     Input('int-threshold', 'value'),
     Input('mz-width', 'value')],
    [State('table-peaklist', 'data')])
def select_peaklist(nc_peaklists, nc_reset, add_row, detect_peaks,
                    clear_peaklist, int_thresh, mz_width, data):
    clear = None
    if len( dash.callback_context.triggered ) == 0:
        set_thresh = False
        set_mzwidth = False
    else:
        clear = ( dash.callback_context.triggered[0]['prop_id'].startswith('B_reset') or 
                  dash.callback_context.triggered[0]['prop_id'].startswith('B_clear_peaklist') )
        add_row = dash.callback_context.triggered[0]['prop_id'].startswith('B_add_peak')   
        set_thresh = dash.callback_context.triggered[0]['prop_id'].startswith('int-threshold') 
        set_mzwidth = dash.callback_context.triggered[0]['prop_id'].startswith('mz-width') 
        detect_peaks = dash.callback_context.triggered[0]['prop_id'].startswith('B_detect_peaks')

    columns = [{"name": i, "id": i, 
                "selectable": True}  for i in PEAKLIST_COLUMNS] 
    
    if add_row: 
        defaults = ['', 0 , 0, 0, 0, 0, 'manual']
        append = {c['id']: d for c,d in zip(columns, defaults)}
        data.append(append)
    
    elif detect_peaks:
        mint.detect_peaks()
        peaklist = mint.peaklist
        data = peaklist.to_dict('records')

    elif set_thresh:
        data = pd.DataFrame(data)
        data['intensity_threshold'] = int_thresh
        data = data.to_dict('records')
    
    elif set_mzwidth:
        data = pd.DataFrame(data)
        data['mz_width'] = mz_width
        data = data.to_dict('records')
    
    elif clear:
        mint.clear_peaklist()
        data = mint.peaklist.to_dict('records')

    elif (nc_peaklists is not None) and (not clear):
        root = Tk()
        root.withdraw()
        root.call('wm', 'attributes', '.', '-topmost', True)
        files = filedialog.askopenfilename(multiple=True)
        files = [i  for i in files if (i.endswith('.csv') or (i.endswith('.xlsx')))]
        if len(files) != 0:
            df = read_peaklists(files)
            data = df.to_dict('records')
        root.destroy()
    
    table = DataTable(
                id='table-peaklist',
                columns=columns,
                data=data,
                sort_action="native",
                sort_mode="single",
                row_selectable=False,
                row_deletable=True,
                editable=True,
                column_selectable=False,
                selected_rows=[],
                page_action="native",
                page_current= 0,
                page_size= 30,
                filter_action='native',                
                style_table={'overflowX': 'scroll'},
                style_as_list_view=True,
                style_cell={'padding-left': '5px', 
                            'padding-right': '5px'},
                style_header={'backgroundColor': 'white',
                              'fontWeight': 'bold'},
                export_format='csv',
                export_headers='display',
                merge_duplicate_headers=True
            )    

    style = {'display': 'inline'}   
    return [table], style 

@app.callback(
    Output('run-mint-div', 'style'),
    [Input('table-peaklist', 'data')]
    )
def show_run_control(table):
    if len(table) == 0:
        style = {'display': 'none'}
    else:
        style = {'display': 'inline'}
    return style

### Button styles
@app.callback(
    [Output('B_peaklists', 'style'),
     Output('B_add_files', 'style'),
     Output('B_run', 'style'),
     Output('B_export', 'style'),
     Output('B_heatmap', 'style'),
     Output('B_shapes', 'style'),
     Output('B_shapes3d', 'style')],
    [Input('B_peaklists', 'n_clicks'),
     Input('B_add_files', 'n_clicks'),
     Input('B_reset', 'n_clicks'),
     Input('progress-bar', 'value')],
    [State('table-peaklist', 'data')])
def run_button_style(nc_peaklists, nc_files, nc_reset, progress, peaklist):
    style_peaklists = button_style('next')
    style_files     = button_style('next')    
    style_run       = button_style('wait')
    style_export    = button_style('wait')
    style_heatmap   = button_style('wait')
    if len(peaklist) == 0: 
        style_peaklists = button_style('next')
    elif len(peaklist) != 0:
        style_peaklists = button_style('ready')
    if mint.n_files != 0:
        style_files= button_style('ready')
        style_run = button_style('next')
    if (mint.n_files != 0) and (mint.n_files != 0) and (progress == 100):
        style_run = button_style('ready')
        style_export = button_style('next')
    if len(mint.results) > 0:
        style_heatmap = button_style('ready')
         
    return (style_peaklists, style_files, style_run, 
            style_export, style_heatmap, style_heatmap, 
            style_heatmap)


### CPU text
@app.callback(
    Output('cpu-text', 'children'),
    [Input('n_cpus', 'value')] )
def mint_cpu_info(value):
    return f'Using {value} cores.'


### Storage
@app.callback(
    Output('storage', 'children'),
    [Input('B_run', 'n_clicks'),
     Input('B_reset', 'value'),
     Input('table-ms-files', 'data'),
     Input('table-peaklist', 'data')],
    [State('n_cpus', 'value'),
     State('storage', 'children'),
    ])
def run_mint(n_clicks, n_clicks_clear, ms_files, peaklist, n_cpus, old_results):
    
    if n_clicks == 0:
        raise PreventUpdate

    ms_files = pd.DataFrame(ms_files)
    if len(ms_files) == 0 :
        files = []
    else:
        files = ms_files['MS-files'].values
        files = [str(P(i)) for i in files]


    peaklist = pd.DataFrame(peaklist)
    if 'peak_label' not in peaklist.columns:
        print('No column "peak_label" in peaklist.')
        raise PreventUpdate

    peaklist = standardize_peaklist(peaklist)    

    if mint.status == 'running' or len(peaklist) == 0 :
        print('MINT status:', mint.status)
        print('Len peaklist:', len(peaklist))
        raise PreventUpdate
    
    #reset = dash.callback_context.triggered[0]['prop_id'].startswith('B_reset')
    run_mint = dash.callback_context.triggered[0]['prop_id'].startswith('B_run')

    #mint = Mint()
    if run_mint:
        print(files)
        mint.peaklist = peaklist
        mint.files = files

    if old_results is not None:
        #old_results = pd.read_json(old_results, orient='split')
        old_results = mint.results

        old_peaklist_features = standardize_peaklist(old_results[PEAKLIST_COLUMNS].drop_duplicates())
        new_peaklist_features = standardize_peaklist(peaklist[PEAKLIST_COLUMNS].drop_duplicates())
    
        diff = diff_peaklist(old_peaklist_features, 
                             new_peaklist_features)

        if len(diff) == 0:
            print('No difference in peaklist')
        else:
            print(diff)

    if run_mint:
        print("Running MINT")
        mint.run(nthreads=n_cpus)
    
    new_results = mint.results

    print('Old results:', old_results)

    # Update results data with new data
    if (old_results is not None) and (len(old_results) > 0):        
        old_results = old_results.set_index(['peak_label', 'ms_file'])
        new_results = mint.results.set_index(['peak_label', 'ms_file'])
        old_results = old_results.drop(new_results.index, errors='ignore')
        new_results = pd.concat([old_results, new_results]).reset_index()

    # Restrict results to files in file list and peak_labels in peaklist
    print('Results lenght before:', len(new_results))
    print('Files:', files)
    print('Files in results:', new_results.ms_file.drop_duplicates().values)
    print('Peak labels:', peaklist.peak_label)
    print('Peak labels in results:', new_results.peak_label.values)

    new_results = new_results[new_results.ms_file.isin(files) & 
                              new_results.peak_label.isin(peaklist.peak_label)]

    #print('Results columns:', new_results.columns)
    print('Results length after:', len(new_results))

    return new_results.to_json(orient='split')


### Data Table
@app.callback(
    [Output('run-out', 'children'),
     Output('peak-select', 'options'),
     Output('peakshapes-selection', 'options'),
     Output('results-div', 'style')],
    [Input('storage', 'children'),
     Input('label-regex', 'value'),
     Input('table-value-select', 'value'),
     Input('B_reload', 'n_clicks')])
def get_table(json, label_regex, col_value, clicks):
    if json is None:
        raise PreventUpdate

    df = pd.read_json(json, orient='split')
    
    # Columns to show in frontend    
    cols = ['ms_file', 'peak_label', 'peak_area', 'peak_n_datapoints', 'peak_max', 'peak_min',
            'peak_median', 'peak_mean', 'file_size', 'peak_delta_int',
            'peak_rt_of_max']     
    
    # Don't update without data
    if len(df) == 0:
        raise PreventUpdate
    
    # Only show file name, not complete path
    if df['ms_file'].apply(basename).value_counts().max() > 1:
        df['ms_file'] = df['ms_file'].apply(basename)
    
        
    # Extract names of biomarkers for other fontend elements
    # before reducing it to frontend version
    biomarkers = df.groupby('peak_label').mean().peak_max.sort_values(ascending=False).index.astype(str)        
    biomarker_options = [ {'label': i, 'value': i} for i in biomarkers]
    
    # Order dataframe columns               
    df = df[cols]
    
    if col_value != 'full':
        df = pd.crosstab(df.peak_label, df.ms_file, df[col_value].astype(np.float64), aggfunc=np.mean).T
        df = df.loc[:, (df != 0).any(axis=0)]  # remove columns with only zeros
        df = remove_all_zero_columns(df)
        df = sort_columns_by_median(df)
        if col_value in ['peakArea']:
            df = df.round(0)
        df.reset_index(inplace=True)
        df.fillna(0, inplace=True)
        
    # Generate labels 
    if (label_regex is not None) and (label_regex != ''):
        labels = [ '.'.join(i.split('.')[:-1]).split('_')[int(label_regex)] for i in df.ms_file ]
        df['Label'] = labels
    else:
        df['Label'] = df.ms_file
    
    df = df.set_index('Label').reset_index()

    df.columns = df.columns.astype(str)
        
    table = DataTable(
                id='table',
                columns=[{"name": i, "id": i, 
                          "selectable": True,
                          "deletable": True} for i in df.columns],
                data=df.to_dict('records'),
                sort_action="native",
                sort_mode="single",
                row_selectable=False,
                row_deletable=False,
                column_selectable=False,
                selected_rows=[],
                page_action="native",
                page_current= 0,
                page_size= 30,
                style_table={'overflowX': 'scroll'},
                style_as_list_view=True,
                style_cell={'padding': '5px'},
                filter_action='native',
                style_header={'backgroundColor': 'white',
                              'fontWeight': 'bold'},
            )
    
    analysis_style = {}
    
    return table, biomarker_options, biomarker_options, analysis_style


# HEATMAP TOOL
@app.callback(
    [Output('heatmap', 'figure'),
     Output('heatmap', 'style')],
    [Input('B_heatmap', 'n_clicks')],
    [State('checklist', 'value'),
     State('table', 'derived_virtual_indices'),
     State('table', 'data'),
     State('table-value-select', 'value')])
def plot_0(n_clicks, options, ndxs, data, column):
    if (n_clicks is None) or (column == 'full') or (len(data) == 0):
        return {}, {'display': 'none'}
    
    numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
    df = pd.DataFrame(data).set_index('Label').select_dtypes(include=numerics)
    
    # Same order as data-table
    df = df.iloc[ndxs]
    print('Heatmap options:', options)
    fig = plot_heatmap(df, 
                       normed_by_cols = 'normed' in options,
                       transposed = 'transposed' in options,
                       correlation = 'correlation' in options,
                       call_show = 'new_tab' in options,
                       add_dendrogram = 'dendrogram' in options,
                       clustered = 'clustered' in options,
                       name=column)    

    if ('new_tab' in options) or (fig is None):
        return {}, {'display': 'none'}
    else:
        return fig, {'min-height': 200, 'width': '100%', 
                     'display': 'inline-block'}


# PEAKEXPLORER
# @lru_cache(maxsize=32)
@app.callback(
    [Output('peakShape', 'figure'),
     Output('peakShape', 'style')],
    [Input('B_shapes', 'n_clicks')],
    [State('n_cols', 'value'),
     State('check_peakShapes', 'value'),
     State('peakshapes-selection', 'value')])
def plot_1(n_clicks, n_cols, options, biomarkers):
    if (len(mint.results) == 0) or (n_clicks == 0):
        return {}, {'display': 'none'}

    if 'legend_horizontal' in options:
        legend_orientation = "h"
    else:
        legend_orientation = 'v'


    fig = plot_peak_shapes(mint.results, n_cols, biomarkers, 
                               legend='legend' in options, 
                               legend_orientation=legend_orientation,
                               call_show='new_tab' in options)

    if ('new_tab' in options) or (fig is None):
        return {}, {'display': 'none'}
    else:
        return fig, {'display': 'inline-block', 'width': '100%'}


@lru_cache(maxsize=32)
@app.callback(
    [Output('peakShape3d', 'figure'),
     Output('peakShape3d', 'style')],
    [Input('B_shapes3d', 'n_clicks')],
    [State('peak-select', 'value'),
     State('check_peakShapes3d', 'value')])
def plot_3d(n_clicks, peak_label, options):
    if (n_clicks is None) or (len(mint.results) == 0) or (peak_label is None):
        return {}, {'display': 'none'}

    if 'legend_horizontal' in options:
        legend_orientation = "h"
    else:
        legend_orientation = 'v'

    fig = plot_peak_shapes_3d(mint.results, peak_label, legend='showlegend' in options, 
                              legend_orientation=legend_orientation, 
                              call_show='new_tab' in options)

    if ('new_tab' in options) or (fig is None):
        return {}, {'display': 'none'}
    else:
        return fig, {'height': 800, 'display': 'inline-block'}


@app.callback(
    Output('heatmap-message', 'children'),
    [Input('table-value-select', 'value')]
)
def return_message(value):
    if value == 'full':
        return 'Heatmap does not work with "Full Table"'
    else:
        return None


## Results Export (Download)
@app.server.route('/export/')
def download_csv():
    file_buffer = io.BytesIO()
    writer = pd.ExcelWriter(file_buffer) #, engine='xlsxwriter')
    mint.peaklist.to_excel(writer, 'Peaklist', index=False)    
    mint.results.to_excel(writer, 'Results Complete', index=False)
    mint.crosstab().T.to_excel(writer, 'PeakArea Summary', index=True)
    meta = pd.DataFrame({'Version': [mint.version], 
                            'Date': [str(date.today())]}).T[0]
    meta.to_excel(writer, 'MetaData', index=True, header=False)
    writer.close()
    file_buffer.seek(0)
    now = datetime.now().strftime("%Y-%m-%d")
    uid = str(uuid.uuid4()).split('-')[-1]
    return send_file(file_buffer,
                     attachment_filename=f'MINT__results_{now}-{uid}.xlsx',
                     as_attachment=True,
                     cache_timeout=0)




import platform

import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import subprocess

from dash_table import DataTable

import numpy as np

from multiprocessing import cpu_count
from ms_mint import __version__
from ms_mint.standards import PEAKLIST_COLUMNS

from .button_style import button_style

n_cpus = cpu_count()

slider_style = {'marginBottom': '3em'}

info_style   =   {'margin-top': 'auto', 
                  'margin-bottom': 10, 
                  'margin-left': 'auto',
                  'display': 'inline', 
                  'float': 'right', 
                  'color': 'grey'
                  }

layout_style = {'max-width': '80%', 
                'margin': 'auto', 
                'margin-bottom': '10%'
                }

space = html.Div(style={'padding': 50})

config={'displaylogo': False}

def get_versions():
    string = ''
    try:
        string += subprocess.getoutput('conda env export --no-build')
    except:
        pass
    return string


ISSUE_TEXT = f'''
%0A%0A%0A%0A%0A%0A%0A%0A%0A
MINT version: {__version__}%0A
OS: {platform.platform()}%0A
Versions:
{get_versions()}
'''


status = html.Div([
    # Status
    html.P(f'Version: {__version__}', style={'text-align': 'right', 'color': 'grey'}),
    html.Div(id='n_peaklist_selected', children=0, style={'display': 'none'}),
    html.Div(id='n_files_selected', children=0, style={'display': 'none'}),
    html.Button('Reset', id='B_reset', style=button_style('warn')),
    html.A(href='https://soerendip.github.io/ms-mint/', 
         children=[html.Button('Help', id='B_help', style=button_style('help', float="right"))],
         target="_blank"),
    html.A(href=f'https://github.com/soerendip/ms-mint/issues/new?body={ISSUE_TEXT}', 
         children=[html.Button('Issues', id='B_issues', style=button_style('help', float="right"))],
         target="_blank"),         
])



ms_files = html.Div([
    # Buttons
    html.Button('Add MS-file(s)', id='B_add_files', style=button_style()),    
    dcc.Checklist(id='files-check', 
                  options=[{ 'label': 'Add files from directory', 'value': 'by-dir' }], 
                  value=['by-dir'], style={'display': 'inline-block'}),
    html.Div(id='files-text', children='', style=info_style),

    dcc.Loading( children=[ html.Div(id='table-ms-files-container', 
                                     style={'min-height':  100, 'margin-top': 10},
                                     children=[
                                         DataTable(id='table-ms-files',
                                                   columns=[ {"name": i, "id": i, "selectable": True}  
                                                             for i in ['MS-files']],
                                                   data=[],
                                                   row_selectable=False,
                                                   row_deletable=True,
                                                   style_cell={'textAlign': 'left'},
                                                   sort_action='native'
                                                   )
                                            ])
                        ]
                ),

    html.Div(style={'padding': 50}),
])



peaklist = html.Div(id='peaklist-div', 
  children=[
    html.H2('Peaklist'),
    html.Button('Import peaklist', id='B_peaklists', style=button_style('')),
    html.Button('Detect peaks', id='B_detect_peaks', style=button_style('')),
    html.Button('Add Peak', id='B_add_peak', style=button_style('')),
    html.Button('Clear', id='B_clear_peaklist', style=button_style('')),

    dcc.Loading( children=[ html.Div(id='table-peaklist-container', 
                                     style={'min-height':  100, 'margin-top': 10},
                                     children=[
                                         DataTable(id='table-peaklist',
                                                   columns=[ {"name": i, "id": i, "selectable": True}  
                                                             for i in PEAKLIST_COLUMNS],
                                                   data=[])
                                            ])
                        ]
                ),
    dcc.Input(id="int-threshold", type='number', placeholder="Intensity Threshold"),
    dcc.Input(id="mz-width", type='number', placeholder="m/z widths"),

    ], style={'display': 'inline'})



run_mint = html.Div(id='run-mint-div', 
  children=[
    html.H2('Run Mint'),
    html.Div(id='peaklist-text', children='', style=info_style),
    html.Div(id='cpu-text', children='', style=info_style),
    html.Br(),
    html.P("Select number of cores:", style={'display': 'inline-block', 'margin-top': 30}),
    html.Div(dcc.Slider( id='n_cpus', 
            min=1, max=n_cpus,
            step=1, value=n_cpus,
            marks={i: f'{i} cpus' for i in [1, n_cpus]}),
        style=slider_style),
    html.Button('Run', id='B_run'),
    html.A(html.Button('Export', id='B_export'), href="export"),
    # Progress bar
    dcc.Interval(id="progress-interval", n_intervals=0, interval=500, disabled=False),
    dbc.Progress(id="progress-bar", value=0),
    html.Div(id='progress', children=[], style=info_style),    
  ], style={'display': 'none'})



results = html.Div(
  id = 'results-div', 
  style = {'visibility': 'hidden'},
  children = [    
    html.H2("Table View", style={'margin-top': 100}),

    dcc.Dropdown(id='table-value-select', value='full',
        options=[ 
            {'label': 'Full table', 'value': 'full'},
            {'label': 'Peak area', 'value': 'peak_area'},
            {'label': 'Retention time of maximum', 'value': 'peak_rt_of_max'},
            {'label': 'N Datapoints', 'value': 'peak_n_datapoints'},
            {'label': 'Peak maximum', 'value': 'peak_max'},
            {'label': 'Peak mininimum', 'value': 'peak_min'},
            {'label': 'Peak median', 'value': 'peak_median'},
            {'label': 'Peak mean', 'value': 'peak_mean'},
            {'label': 'First minus last intensity', 'value': 'peak_delta_int'}
        ]),
    
    dcc.Loading( children=[ html.Div(id='run-out', 
                                     style={'min-height':  0, 'margin-top': 10},
                                     children=[
                                        DataTable( id='table', data=np.array([]) )
                                        ]
                                    )
                        ]
                ),

    dcc.Input(id="label-regex", type='text', placeholder="Label Selector"),
    
    html.Button('Reload', id='B_reload'),
    
    html.Div(style={'padding': 50}),
        
    html.H2("Peak Shapes"),
    
    dcc.Dropdown(id='peakshapes-selection',
        options=[],
        value=[],
        multi=True
        ), 
    
    html.Button('Peak Shapes', id='B_shapes', style=button_style()),
    
    dcc.Checklist(id='check_peakShapes', 
        options=[{'label': 'Show legend', 'value': 'legend'},
                 {'label': 'Horizontal legend', 'value': 'legend_horizontal'},
                 {'label': 'Show in new tab', 'value': 'new_tab'}],
        value=['legend'], style={'display': 'inline-block'}),
    
    html.Div(dcc.Slider(id='n_cols', min=1, max=12, step=1, value=2,
            marks={i: f'{i} columns' for i in range(1, 13)}),
        style=slider_style),
    
    dcc.Loading( children=[ dcc.Graph(id='peakShape', figure={}, config=config) ]),
    
    html.Div(style={'padding': 50}),
    
    html.H2("Heatmap"),
    
    html.Button('Heatmap', id='B_heatmap', style=button_style()),
    
    dcc.Checklist(id='checklist', 
        options=[
            { 'label': 'Normalized by biomarker', 'value': 'normed'},
            { 'label': 'Cluster', 'value': 'clustered'},
            { 'label': 'Dendrogram', 'value': 'dendrogram'},
            { 'label': 'Transposed', 'value': 'transposed'},
            { 'label': 'Correlation', 'value': 'correlation'},
            { 'label': 'Show in new tab', 'value': 'new_tab'}],

        value=['normed'], style={'display': 'inline-block'}),

    html.P(id='heatmap-message'),
    
    dcc.Loading(children=[ dcc.Graph(id='heatmap', figure={}, config=config) ]),
                
    html.Div(style={'padding': 50}),

    html.H2("Peak Shapes 3D"),
    
    html.Button('Peak Shapes 3D', id='B_shapes3d', style=button_style()),
    
    dcc.Checklist(id='check_peakShapes3d', 
        options=[{'label': 'Show legend', 'value': 'legend'},
                 {'label': 'Horizontal legend', 'value': 'legend_horizontal'},
                 {'label': 'Show in new tab', 'value': 'new_tab'}],
        value=['legend'], style={'display': 'inline-block'}),
    
    dcc.Dropdown(id='peak-select', options=[]),
    dcc.Loading([ dcc.Graph(id='peakShape3d', figure={}, config=config) ])
  ])

Layout = html.Div(
  [   
    html.H1('MINT', style={'margin-top': '10%'}),
    html.Div(id='storage', style={'display': 'none'}),
    status,   space,
    ms_files, space,
    peaklist, space,
    run_mint, space,
    results
    
], style=layout_style)
